#ifndef TSERIES_HH
#define TSERIES_HH

#include <string>
#include <iosfwd>
#include <memory>
#include "Complex.hh"
#include "Interval.hh"
#include "Time.hh"

class DVector;
class FSeries;
class Chirp;

/** The TSeries class holds a time series of arbitrarily typed data.
  * @memo Time series API.
  * @author John G. Zweizig
  * @version 1.4; Last modified Aug 9, 2006
  */
class TSeries {
public:
  /**  Vector length data type.
    *  @memo Length type;
    */
  typedef unsigned long size_type;

  /**  Data Status Type.
    *  @memo Status type;
    */
  typedef unsigned long stat_type;

  /**  Mathematical data type.
    *  @memo Math type;
    */
  typedef double math_type;

  /**  A time series object is initialized without a data vector.
    *  @memo Default constructor.
    */
  TSeries();

  /**  The argument series is duplicated into the new TSeries.
    *  @memo  Copy Constructor.
    *  @param TS Time series to be duplicated.
    */
  TSeries(const TSeries& TS);

  /**  A time series is created with a specified start time and sample 
    *  interval, but no data.
    *  @memo  Dataless constructor.
    *  @param t0 Series start time.
    *  @param dt Step interval.
    */
  TSeries(const Time& t0, const Interval& dt);

  /**  A float time series object is constructed and initialized from a 
    *  float data array.
    *  @memo  Float data constructor.
    *  @param t0    Series start time.
    *  @param dt    Time interval between two successive samples.
    *  @param NData Number of data words.
    *  @param dData float array containing series data.
    */
  TSeries(const Time& t0, const Interval& dt, size_type NData, 
	  const float dData[]=0);

  /**  A short time series object is constructed and initialized from a 
    *  short data array.
    *  @memo  Short integer data constructor.
    *  @param t0    Series start time.
    *  @param dt    Time interval between two successive samples.
    *  @param NData Number of data words.
    *  @param dData short integer array containing series data.
    */
  TSeries(const Time& t0, const Interval& dt, size_type NData, 
	  const short dData[]);

  /**  An integer time series object is constructed and initialized from a 
    *  integer data array.
    *  @memo  Integer data constructor.
    *  @param t0    Series start time.
    *  @param dt    Time interval between two successive samples.
    *  @param NData Number of data words.
    *  @param dData float array containing series data.
    */
  TSeries(const Time& t0, const Interval& dt, size_type NData, 
	  const int dData[]);

  /**  A double float time series object is constructed and initialized from 
    *  a double float data array.
    *  @memo  Double float data constructor.
    *  @param t0    Series start time.
    *  @param dt    Time interval between two successive samples.
    *  @param NData Number of data words.
    *  @param dData double array containing series data.
    */
  TSeries(const Time& t0, const Interval& dt, size_type NData, 
	  const double dData[]);

  /**  A short time series object is constructed and initialized from a 
    *  Complex data array.
    *  @memo  Complex data constructor.
    *  @param t0    Series start time.
    *  @param dt    Time interval between two successive samples.
    *  @param NData Number of data words.
    *  @param dData Complex array containing series data.
    */
  TSeries(const Time& t0, const Interval& dt, size_type NData, 
	  const fComplex dData[]);

  /**  Construct a time series and initialize it from a specified signal
    *  template function.
    *  @memo  Signal template function constructor.
    *  @param t0    Series start time.
    *  @param dt    Time interval between two successive samples.
    *  @param NData Number of data words.
    *  @param func  Template object based on a Chirp.
    */
  TSeries(const Time& t0, const Interval& dt, size_type NData, 
	  const Chirp& func);

  /**  Construct a time series and initialize it from a specified 
    *  data vector.
    *  @memo  Signal template function constructor.
    *  @param t0    Series start time.
    *  @param dt    Time interval between two successive samples.
    *  @param data  Data vector.
    */
  TSeries(const Time& t0, const Interval& dt, const DVector& data);

  /**  Construct a time series and initialize it from a specified
    *  data vector.
    *  @memo  Signal template function constructor.
    *  @param t0    Series start time.
    *  @param dt    Time interval between two successive samples.
    *  @param data  Data vector.
    */
  TSeries(const Time& t0, const Interval& dt, DVector* data);

  /**  Construct a time series from an FSeries by inverse fft.
    *  @memo  FSeries constructor.
    *  @param fs    Frequency Series.
    */
  explicit TSeries(const FSeries& fs);

  /**  Blow away the DVector.
    *  @memo TSeries Destructor.
    */
  virtual ~TSeries();

  //------------------------------  Accessors
  /**  Convert and copy the argument TSeries to this TSeries using the 
    *  existing DVector.
    *  @memo   Copy TSeries data.
    *  @param  in  Time series to be copied.
    *  @return reference to this time series.
    */
  TSeries& copy(const TSeries& in);

  /**  Write a formatted dump of the series contents to the specified output
    *  stream.
    *  @memo   Formatted dump of the TSeries data.
    *  @param  out  Output stream to receive the formatted TSeries dump.
    *  @return reference to the specified output stream.
    */
  std::ostream& Dump(std::ostream& out) const;

  /**  A TSeries is decimated by the specified integer factor. The resulting
    *  series is returned. The first sample in the TSeries is included in the 
    *  decimated TSeries.
    *  @memo   Decimate a Time Series.
    *  @param  dec Decimation factor.
    *  @return A decimated time series.
    */
  TSeries decimate(size_type dec) const;

  /**  Remove the data from the time interval #\{t0, t0+dT\}#. The time 
    *  removed is rounded to the nearest integer number of samples.
    *  @memo Erase data from the start of a TSeries.
    *  @param  dT Time duration of the substring to be erased.
    */
  void eraseStart(Interval dT);

  /**  A TSeries is returned containing the data from the time interval
    *  t0 - t0+dT. If dT == 0, all data after t0 will be returned.
    *  @memo Extract a subset of the data from a TSeries.
    *  @return a TSeries containing the requested subset of the data.
    *  @param  t0 Start time of the desired subset.
    *  @param  dT Time duration of the desired subset.
    */
  TSeries extract(const Time& t0, const Interval& dT=Interval(0.0)) const;

  /**  The series is extended to the specified time by padding on the right 
    *  with zeroes.
    *  @memo Extend a TSeries.
    *  @param  t0 Desired series end-time.
    */
  void extend(const Time& t0);

  /**  Get the bin index corresponding to the specified time. Note that the
    *  Time of the indexed sample is less than or equal to the specified time.
    *  @memo   Get bin index.
    *  @return Index of the bin with time t.
    */
  size_type getBin(const Time& t) const;

  /**  Get the time corresponding to the specified bin number.
    *  @memo   Get bin time.
    *  @return Time corresponding to the specified bin.
    */
  Time getBinT(size_type inx) const;

  /**  The specified entry is converted to complex if necessary
    *  and returned.
    *  @memo   Get ith element as a complex number.
    *  @param  index Index of the element to be returned.
    *  @return ith element value.
    */
  fComplex getComplex(size_type index) const;

  /**  The TSeries data are copied to a short int data buffer pointed to 
    *  by 'data'. The data will be converted to 'short int's if the data 
    *  vector is of a different data type. A maximum of 'len' data words 
    *  will be copied to the data buffer. 
    *  @memo   Copy data to a short int output array.
    *  @param  len  Maximum number of words to copy.
    *  @param  data Short integer array into which the data will be copied.
    *  @return The number of words copied.
    */
  size_type getData(size_type len, short* data) const;

  /**  The TSeries data are copied to an integer data buffer pointed to by
    *  'data'. The data will be converted to floats if the data vector
    *  is of a different data type. A maximum of 'len' data words will 
    *  be copied to the data buffer. getData() returns the number of 
    *  words copied.
    *  @memo   Copy data to an integer array
    *  @param  len  Dimension of output array.
    *  @param  data Array to receive series data.
    *  @return Number of words copied.
    */
  size_type getData(size_type len, int* data) const;

  /**  The TSeries data are copied to a float data buffer pointed to by
    *  'data'. The data will be converted to floats if the data vector
    *  is of a different data type. A maximum of 'len' data words will 
    *  be copied to the data buffer. getData() returns the number of 
    *  words copied.
    *  @memo   Copy data to a float array
    *  @param  len  Dimension of output array.
    *  @param  data Array to receive series data.
    *  @return Number of words copied.
    */
  size_type getData(size_type len, float* data) const;

  /**  The TSeries data are copied to a double data buffer pointed to by
    *  'data'. The data will be converted to floats if the data vector
    *  is of a different data type. A maximum of 'len' data words will 
    *  be copied to the data buffer. getData() returns the number of 
    *  words copied.
    *  @memo   Copy data to a double array
    *  @param  len  Dimension of output array.
    *  @param  data Array to receive series data.
    *  @return Number of words copied.
    */
  size_type getData(size_type len, double* data) const;

  /**  The TSeries data are copied to a Complex buffer pointed to by
    *  'data'. The data will be converted to fComplex if the data vector
    *  is of a different data type. A maximum of 'len' data words will 
    *  be copied to the data buffer. getData() returns the number of 
    *  words copied.
    *  @memo   Copy data to a complex array.
    *  @param  len  Dimension of output array
    *  @param  data Array to receive series data.
    *  @return Number of words copied.
    */
  size_type getData(size_type len, fComplex* data) const;

  /**  The specified entry is converted to double precision if necessary
    *  and returned. If the series is complex, the real part of the 
    *  indicated element is returned.
    *  @memo   Get ith element as a double precision number.
    *  @param  index Index of the element to be returned.
    *  @return ith element value.
    */
  double getDouble(size_type index) const;

  /**  The GPS time of the last plus one is returned as a Time object.
    *  @memo Get the end Time of the series.
    *  @return End time of the series.
    */
  Time getEndTime(void) const;

  /**  getF0 returns the heterodyne frequency used if the time series
    *  was frequency shifted.
    *  @memo   Get heterodyne frequency.
    *  @return Frequency used to down-convert frequencies.
    */
  double getF0(void) const {return mF0;}

  /**  getFNyquist returns the positive Nyquist frequency of the TSeries.
    *  @memo   Get the Nyquist frequency.
    *  @return Maximum meaningful frequency.
    */
  double getFNyquist(void) const {return mFNyquist;}

  /**  Returns the total time duration of the time series.
    *  @memo   Get series duration.
    *  @return Time duration of the series.
    */
  Interval getInterval(void) const;

  /**  Get a valid pointer to a zero-terminated string containing the series 
    *  name. If the name is undefined a pointer to a null string is returned.
    *  @memo   Get the series name.
    *  @return a constant pointer to the series name string.
    */
  const char* getName(void) const;

  /**  Returns the number of samples stored in the data vector.
    *  @memo   Get the number of samples
    *  @return the number of samples.
    */
  size_type getNSample(void) const;

  /**  Get the GPS time of the first sample represented as a Time object.
    *  @memo Get the start time.
    *  @return Time of first sample.
    */
  Time getStartTime(void) const;

  /**  The time bin width (the inverse of the sampling frequency) is 
    *  returned as an Interval object.
    *  @memo Get the sample time.
    *  @return the sample time interval.
    */
  Interval getTStep(void) const;

  /**  The average value of the data sequence is calculated. Zero is returned
    *  if the series is empty.
    *  @memo Average value of data.
    *  @return Average data value.
    */
  math_type getAverage(void) const;

  /**  The average value of the data sequence is calculated. Series elements
    *  are converted to compelx if necessary before being averaged.  Zero is 
    *  returned if the series is empty.
    *  @memo Average value of data.
    *  @return Average data value.
    */
  fComplex getComplexAverage(void) const;

  /**  The maximum value of the data sequence is found. Zero is returned
    *  if the series is empty.
    *  @memo Maximum data value.
    *  @return Maximum data value.
    */
  math_type getMaximum(void) const;

  /**  The minimum value in the sequence is found. Zero is returned
    *  if the series is empty.
    *  @memo Minumim data value.
    *  @return Minimum data value.
    */
  math_type getMinimum(void) const;

  /**  Count the number of entries in the range #low <= x < high#.
    *  @memo Number of entries betwee two limits.
    *  @return Number of entries greater than limit.
    *  @param low  Lower limit.
    *  @param high Upper limit.
    */
  size_type getNBetween(math_type low, math_type high) const;

  /**  Count the number of entries greater than the specified value.
    *  @memo Entries greater than limit.
    *  @return Number of entries greater than limit.
    *  @param low  Lower limit.
    */
  size_type getNGreater(math_type Limit) const;

  /**  Count the number of entries less than the specified value.
    *  @memo Entries less than Limit.
    *  @return Number of entries less than limit.
    */
  size_type getNLess(math_type Limit) const;

  /**  Get the effective bandwidth ratio imposed by any windowing.
    *  @memo Window sigma.
    *  @return Window sigma.
    */
  math_type getSigmaW(void) const;

  /**  Get the status information word.
    *  @memo Data status.
    *  @return Data status word.
    *  @param high Upper limit.
    */
  stat_type getStatus(void) const;

  /**  Test if TSeries is empty.
    *  @memo Test for data.
    *  @return true if no data are available.
    */
  bool isComplex(void) const;

  /**  Test if TSeries is empty.
    *  @memo Test for data.
    *  @return true if no data are available.
    */
  bool empty(void) const;
  bool isEmpty(void) const {return empty();}

  /**  Test the data status. The data are considered good if the status
    *  word is zero.
    *  @memo Test data status.
    *  @return True if status indicates data are good.
    */
  bool isGood(void) const;

  /**  Get a constant pointer to the data vector.
    *  @memo Get a reference to the data vector.
    *  @return a constant pointer to the data vector.
    */
  const DVector* refDVect(void) const {return mData.get();}

  /**  Get a constant pointer to the data.
    *  @memo   Get a constant reference to the data.
    *  @return a pointer to the data.
    */
  const void* refData(void) const;

  /**  Get a pointer to the data.
    *  @memo   Get a reference to the data.
    *  @return a pointer to the data.
    */
  void* refData(void);

  //------------------------------  Mutators
  /**  The specified character string is appended to the time series name 
    *  string.
    *  @memo   Append a string to the series name.
    *  @param  name Constant pointer to a character array to be appended 
    *               to the series name.
    */
  void appName(const char* name);

  /**  Combine the data status word with the specified value.
    *  @memo Set the data status.
    *  @param stat Status value.
    */
  void combineStatus(stat_type stat);

  /**  The supplied data are converted if necessary to the vector data 
    *  type and copied to the series vector. If the vector is not defined, 
    *  a vector of short integers will be created.
    *  @memo   Set the series data.
    *  @param  t    New series start time.
    *  @param  Step New series sample interval.
    *  @param  data Short integer array containing new sample
    *  @param  N    Number of data words in 'data'
    *  @return 0 if successful.
    */
  int setData(const Time& t, const Interval& Step, const short* data, 
		 size_type N);

  /**  The supplied data are converted if necessary to the vector data 
    *  type and copied to the series vector. If the vector is not defined, 
    *  a vector of floats will be created.
    *  @memo   Set the series data.
    *  @param  t    New series start time.
    *  @param  Step New series sample interval.
    *  @param  data Float array containing new sample
    *  @param  N    Number of data words in 'data'
    *  @return 0 if successful.
    */
  int setData(const Time& t, const Interval& Step, const int* data, 
		 size_type N);

  /**  The supplied data are converted if necessary to the vector data 
    *  type and copied to the series vector. If the vector is not defined, 
    *  a vector of floats will be created.
    *  @memo   Set the series data.
    *  @param  t    New series start time.
    *  @param  Step New series sample interval.
    *  @param  data Float array containing new sample
    *  @param  N    Number of data words in 'data'
    *  @return 0 if successful.
    */
  int setData(const Time& t, const Interval& Step, const float* data, 
		 size_type N);

  /**  The supplied data are converted if necessary to the vector data 
    *  type and copied to the series vector. If the vector is not defined, 
    *  a vector of floats will be created.
    *  @memo   Set the series data.
    *  @param  t    New series start time.
    *  @param  Step New series sample interval.
    *  @param  data Float array containing new sample
    *  @param  N    Number of data words in 'data'
    *  @return 0 if successful.
    */
  int setData(const Time& t, const Interval& Step, const double* data, 
		 size_type N);

  /**  The supplied data are converted if necessary to the vector data 
    *  type and copied to the series vector. If the vector is not defined, 
    *  a vector of fComplex will be created.
    *  @memo   Set the series data.
    *  @param  t    New series start time.
    *  @param  Step New series sample interval.
    *  @param  data fComplex array containing new sample
    *  @param  N    Number of data words in 'data'
    *  @return 0 if successful.
    */
  int setData(const Time& t, const Interval& Step, const fComplex* data, 
		 size_type N);

  /**  Store the specified frequency in the TSeries central frequency field.
    *  The data sre left unmodified. The central frequency is defined as the 
    *  true frequency of the bin that is apparently DC.
    *  @memo  Set the heterodyne frequency.
    *  @param f0 Frequency to which the TSeries heterodyne frequency will be set.
    */
  void setF0(double f0);

  /**  Define the Nyquist frequency of the TSeries. The nyquist frequency is 
    *  set to 2/dT in the constructors and must be set as appropriate by any
    *  decimation filter.
    *  @memo  Set the Nyquist frequency.
    *  @param fNy Nyquist frequency of the TSeries.
    */
  void setFNyquist(double fNy);

  /**  The specified name is copied to the series name string.
    *  @memo Name a series.
    *  @param constant pointer to series name character array.
    */
  void setName(const char* name);

 /**  Set the window effective bin width. This is the RMS of any windowing
    *  function used, i.e. sw = sqrt( Sum(i=0, N ; w(i)^2)/N) .
    *  @memo Set the window sigma.
    *  @param sw Window RMS.
    */
  void setSigmaW(math_type sw);

  /**  Set the data status word to the specified value.
    *  @memo Set the data status.
    *  @param stat Status value.
    */
  void setStatus(stat_type stat);

  /**  The specified data are converted if necessary to the series
    *  data type, and appended to the series. If data already exist in 
    *  the vector, the end time of the existing data must match the start
    *  time of the new data. The step-size of the data to be 
    *  appended must also match that of the existing data.
    *  @memo   Append data to the time series.
    *  @param  t    Start time of data to be appended to the series.
    *  @param  Step sample interval of data to be appended to the series.
    *  @param  data Short integer array containing the data to be appended.
    *  @param  N    Number of data words to be appended to the series.
    *  @return \begin{tabular}{rl} 0 & successful \\
    *          -1 & start time isn't equal to the series end time \\
    *          -2 & sample intervals aren't equal \end{tabular}.
    */
  int Append(const Time& t, const Interval& Step, const short* data, 
	      size_type N);

  /**  The specified data are converted if necessary to the series
    *  data type, and appended to the series. If data already exist in 
    *  the vector, the end time of the existing data must match the start
    *  time of the new data. The step-size of the data to be 
    *  appended must also match that of the existing data.
    *  @memo Append data to the time series.
    *  @param t    Start time of data to be appended to the series.
    *  @param Step sample interval of data to be appended to the series.
    *  @param data Float array containing the data to be appended.
    *  @param N    Number of data words to be appended to the series.
    *  @return \begin{tabular}{rl} 
    *            0 & successful \\ 
    *           -1 & start time isn't equal to the current end time \\
    *           -2 & sample intervals aren't equal. \end{tabular}
    */
  int Append(const Time& TS, const Interval& Step, const int* data, 
	     size_type N);

  /**  The specified data are converted if necessary to the series
    *  data type, and appended to the series. If data already exist in 
    *  the vector, the end time of the existing data must match the start
    *  time of the new data. The step-size of the data to be 
    *  appended must also match that of the existing data.
    *  @memo Append data to the time series.
    *  @param t    Start time of data to be appended to the series.
    *  @param Step sample interval of data to be appended to the series.
    *  @param data Float array containing the data to be appended.
    *  @param N    Number of data words to be appended to the series.
    *  @return \begin{tabular}{rl} 
    *            0 & successful \\ 
    *           -1 & start time isn't equal to the current end time \\
    *           -2 & sample intervals aren't equal. \end{tabular}
    */
  int Append(const Time& TS, const Interval& Step, const float* data, 
	     size_type N);

  /**  The specified data are converted if necessary to the series
    *  data type, and appended to the series. If data already exist in 
    *  the vector, the end time of the existing data must match the start
    *  time of the new data. The step-size of the data to be 
    *  appended must also match that of the existing data.
    *  @memo Append data to the time series.
    *  @param t    Start time of data to be appended to the series.
    *  @param Step sample interval of data to be appended to the series.
    *  @param data Float array containing the data to be appended.
    *  @param N    Number of data words to be appended to the series.
    *  @return \begin{tabular}{rl} 
    *            0 & successful \\ 
    *           -1 & start time isn't equal to the current end time \\
    *           -2 & sample intervals aren't equal. \end{tabular}
    */
  int Append(const Time& TS, const Interval& Step, const double* data, 
	     size_type N);

  /**  The specified data are converted if necessary to the series
    *  data type, and appended to the series. If data already exist in 
    *  the vector, the end time of the existing data must match the start
    *  time of the new data. The step-size of the data to be 
    *  appended must also match that of the existing data.
    *  @memo Append data to the time series.
    *  @param t    Start time of data to be appended to the series.
    *  @param Step sample interval of data to be appended to the series.
    *  @param data Complex array containing the data to be appended.
    *  @param N    Number of data words to be appended to the series.
    *  @return \begin{tabular}{rl} 
    *            0 & successful \\ 
    *           -1 & start time isn't equal to the current end time \\
    *           -2 & sample intervals aren't equal. \end{tabular}
    */
  int Append(const Time& TS, const Interval& Step, const fComplex* data, 
	     size_type N);

  /**  The argument TSeries is optionally decimated and appended to the 
    *  series. The starting time of the argument series must be equal to 
    *  the end time of the existing data and the sample rate must be the
    *  same as the object sample rate. Decimation is not yet implemented.
    *  @memo   Concatenate two TSeries.
    *  @param  TS	time series to be appended.
    *  @param  decim8	decimation factor.
    *  @return \begin{tabular}{rl} 0 & successful \\
    *          -1 & start time of TS not equal to the end time \\
    *          -2 & decimated sample rate isn't equal to the current rate.
    * \end{tabular}
    */
  int Append(const TSeries& t, size_type decim8=1);

  /**  The data vector is converted to the specified type. Valid types
    *  are those enumerated in DVector::DVType.
    *  @memo Convert the data vector type. 
    *  @param type The type to which the series data are to be converted.
    */
  void Convert(int type);

  /**  The data vector is extended to hold the specified number of data 
    *  words. No action is taken if the vector can already hold the 
    *  specified number of data words. The current data are copied if 
    *  vector is reallocated.
    *  @memo Expand the data vector. 
    *  @param len Desired vector capacity in words.
    */
  void ReSize(size_type len);

  /**  The current data vector length is set to zero. The data vector 
    *  remains allocated with the original storage capacity. The start
    *  time and sample interval may be reset or are placed in a "don't
    *  care" state. The series name is left unchanged.
    *  @memo Delete the series data.
    *  @param t0 New start time. By default the time is zeroed.
    *  @param dt New sample interval. By default the interval is zeroed.
    */
  void Clear(const Time& t0=Time(0), const Interval& dt=Interval(0.0));

  /**  fShift up-converts (heterodynes) the time series by the specified
    *  frequency shift. The resulting data vector will be of fComplex type.
    *  The sign of the frequency shift is such that the DC bin is converted 
    *  to the shift frequency. The TSeries f0 value is decreasedby an equal 
    *  amount.
    *  @memo   Up-convert (heterodyne) a time series.
    *  @param  f0 heterodyne frequency.
    *  @param  phi0 Phase shift of the first sample (defaults to 0).
    *  @return A complex TSeries with the heterodyned data.
    */
  TSeries fShift(math_type f0, math_type phi0=0) const;

  /**  The TSeries data are copied to the current TSeries. The current 
    *  TSeries type and name remain unchanged.
    *  @memo   Assignment operator.
    *  @param  TS Constant TSeries to be copied.
    *  @return reference to the updated TSeries.
    */
  TSeries& operator=(const TSeries& ts);

  /**  Take a dot product of the two data vectors.
    *  @memo Correlate two series.
    *  @return Correlation coefficient ($ \Sigma x_i y_i^* $).
    */
  math_type operator*(const TSeries& y) const;

  /**  Divide two series on an element-by-element basis.
    *  @memo   Divide series by another.
    *  @return Float seies containing the ratio of the arguments.
    *  @param  Divisor series.
    */
  TSeries operator/(const TSeries& y) const;

  /**  The argument value is converted to the TSeries data type and added 
    *  to each element of the time series.
    *  @memo   Bias a TSeries.
    *  @param  bias Value added to each element.
    *  @return reference to the modified TSeries.
    */
  TSeries& operator+=(math_type bias);

  /**  The argument value is converted to the TSeries data type and subtracted 
    *  from each element of the time series.
    *  @memo   Bias a TSeries.
    *  @param  bias Value subtracted from each element.
    *  @return reference to the modified TSeries.
    */
  TSeries& operator-=(math_type bias);

  /**  Each element of the time series is multiplied by the rhs value.
    *  The result is converted to the series data type and replaces the
    *  original series.
    *  @memo   Scale a TSeries.
    *  @param  scale Value used to multiply each element.
    *  @return reference to the modified TSeries.
    */
  TSeries& operator*=(math_type scale);

  /**  Each element of the time series is divided by the rhs value.
    *  The result is converted to the series data type and replaces the
    *  original series.
    *  @memo   Scale a TSeries.
    *  @param  scale Value used to divide each element.
    *  @return reference to the modified TSeries.
    */
  TSeries& operator/=(math_type scale);

  /**  Each element of the time series is incremented by the corresponding
    *  element of the argument. The #rhs# argument must have the same sampling
    *  rate and contain a subset of the times in the series being added to.
    *  If the sample rates differ or the #rhs# time bins don't correspond 
    *  exactly to existing time bins, a #std::runtime_error# exception is \
    *  thrown.
    *  @memo   Add a TSeries.
    *  @param  rhs TSeries to be added to the current TSeries.
    *  @return reference to the modified TSeries.
    */
  TSeries& operator+=(const TSeries& rhs);

  /**  Each element of the time series is decremented by the corresponding
    *  elements of the argument. The #rhs# argument must have the same sampling
    *  rate and contain a subset of the times in the series being added to.
    *  If the sample rates differ or the #rhs# time bins don't correspond 
    *  exactly to existing time bins, a #std::runtime_error# exception is 
    *  thrown.
    *  @memo   Subtract a TSeries.
    *  @param  rhs TSeries to be subtracted from the current TSeries.
    *  @return reference to the difference TSeries.
    */
  TSeries& operator-=(const TSeries& rhs);

  /**  Get a pointer to the data vector.
    *  @memo   Get a reference to the data vector.
    *  @return a pointer to the data vector.
    */
  DVector* refDVect(void) {return mData.get();}

  /**  Test for consistency with end of series.
    */
  int Contig(const Time& t, const Interval& Step);

  /**  Overlap finds the ranges of two time series that overlap in time
   */
  size_type overlap(const TSeries& ts, size_type& inx, size_type& inx1) const;

private:
  /**  Series Name
    */
  std::string   mName;

  /**  GPS time of the first sample in the series. If the series represents
    *  a physical quantity integrated over a time bin, the first time bin
    *  is from T0 - T0+dT.
    *  @memo Starting absolute time.
    */
  Time     mT0;

  /**  Time increment between successive samples in the time series.
    *  @memo Time increment or bin size.
    */
  Interval mDt;

  /**  Frequency shift if the series source was heterodyned.
    *  @memo Frequency shift.
    */
  double mF0;

  /**  Nyquist frequency of TSeries.
    *  @memo Nyquist requency.
    */
  double mFNyquist;

  /**  The data status word defaults to zero when the TSeries is constructed
    *  and may be set to a non-zero value to indicate dubious data quality. 
    *  @memo Data status word.
    */
  stat_type mStatus;

  /**  Data vector pointer.
    */ 
  std::auto_ptr<DVector> mData;

  /**  Effective bin width from windowing.
    */
  math_type mSigmaW;
};

#ifndef __CINT__
//--------------------------------------  Get TSeries status
inline bool
TSeries::empty(void) const {
    return (getNSample() == 0);
}

inline Time 
TSeries::getBinT(size_type inx) const {
    Time r(mT0);
    r += mDt * double(inx);
    return r;
}

inline Time
TSeries::getEndTime(void) const {
    return getBinT(getNSample());
}

inline Interval
TSeries::getInterval(void) const {
    return mDt*double(getNSample());
}

inline const char*
TSeries::getName(void) const {
    return mName.c_str();
}

inline Time
TSeries::getStartTime(void) const {
    return mT0;
}

inline Interval
TSeries::getTStep(void) const {
    return mDt;
}

inline TSeries::math_type
TSeries::getSigmaW(void) const {
    return mSigmaW;
}

inline TSeries::stat_type
TSeries::getStatus(void) const {
    return mStatus;
}

inline bool
TSeries::isGood(void) const {
    return (getStatus() == 0);
}

//--------------------------------------  Set the series data
inline int
TSeries::setData(const Time& t, const Interval& Step, const short* data, 
		 size_type N) {
    Clear();
    return Append(t, Step, data, N);
}

inline int
TSeries::setData(const Time& t, const Interval& Step, const int* data, 
		 size_type N) {
    Clear();
    return Append(t, Step, data, N);
}

inline int
TSeries::setData(const Time& t, const Interval& Step, const float* data, 
		 size_type N) {
    Clear();
    return Append(t, Step, data, N);
}

inline int
TSeries::setData(const Time& t, const Interval& Step, const double* data, 
		 size_type N) {
    Clear();
    return Append(t, Step, data, N);
}

inline int
TSeries::setData(const Time& t, const Interval& Step, const fComplex* data, 
		 size_type N) {
    Clear();
    return Append(t, Step, data, N);
}

//--------------------------------------  Bias a series.
inline TSeries&
TSeries::operator-=(math_type bias) {
    return operator+=(-bias);
}

//--------------------------------------  Scale a series.
inline TSeries&
TSeries::operator/=(math_type scale) {
    return operator*=(1.0/scale);
}

#endif // __CINT__

/** @name TSeries functions
  * This section contains functions of TSeries that are not members of the 
  * TSeries class. 
  * @memo Non-member functions of TSeries.
  * @author John G. Zweizig
  * @version 1.1; Last modified October 4, 1999
  */
//@{

/**  Difference of two TSeries.
  *  @memo Difference of two TSeries.
  *  @return TSeries difference between the two arguments.
  *  @param t1 Starting series
  *  @param t2 Series to be subtracted from #t1#.
  */
inline TSeries
operator-(const TSeries& t1, const TSeries& t2) {
    TSeries r(t1);
    r -= t2;
    return r;
}

/**  Sum of two TSeries.
  *  @memo Sum of two TSeries.
  *  @return TSeries sum of the two arguments.
  *  @param t1 Starting series
  *  @param t2 Series to be added to #t1#.
  */
inline TSeries
operator+(const TSeries& t1, const TSeries& t2) {
    TSeries r(t1);
    r += t2;
    return r;
}
//@}
#endif   //  TSERIES_HH
