#ifndef ENGINE_HH
#define ENGINE_HH

#include "eval_stack.hh"
#include <string>
#include <vector>
#include <iosfwd>

/**  Symbolic arithmetic evaluation engine. 
  *  The Engine class parses a symbolic expression and stores the parsed 
  *  Expression as a reverse polish list. The parsed expression can then 
  *  be evaluated an arbitrary number of times. The result(s) are left
  *  on the evaluation stack. Symbol replacement, Function evaluation, and
  *  operator evaluaton may be spcified by replaceent of virtual functions.
  *  The operations implemented in the default version are specified by the
  *  OpsEnum enumerator. The symbols, values and the evaluation order
  *  are listed below.
  *  \begin{tablular}{|l|l|l|l|l|l|l|l|l|}
  *  #+#  & kAdd    & 2 & #-#  & kSub   & 2 & #*#  & kMpy    & 1 \\
  *  #/#  & kDiv    & 1 & #|#  & kOr    & 2 & #&#  & kAnd    & 2 \\
  *  #&&# & kLogAnd & 4 & #||# & kLogOr & 4 & #==# & kCmpEq  & 3 \\
  *  #!=# & kCmpNE  & 3 & #<#  & kCmpLt & 3 & #<=# & kCmpLE  & 3 \\
  *  #>#  & kCmpGt  & 3 & #>=# & kCmpGE & 3 & #%#  & kModulo & 1 \\
  *  \end{tabular}
  *  @author J. Zweizig
  *  @version 1.0; Last Modified July 6, 2004
  */
class Engine {
public:
  /** Construct an evaluation engine.
    */
  Engine(eval_stack& stk);

  /** Destroy the evaluation engine.
    */
  virtual ~Engine(void);

  /** Parse the equation.
    */
  virtual void Parse(std::istream& in);

  /**  Evaluate an already parsed string.
    */
  virtual void crank(void);

  /**  Push the data from the names symbol
    */
  virtual void push_symbol(const std::string& sym);

  /**  Push the specified constant value onto the evaluation stack.
   */
  virtual void push_numeric(const std::string& lit);

  /**  Push the specified constant value onto the evaluation stack.
   */
  virtual void push_literal(const std::string& lit);

  /**  Evaluate a specified function.
   */
  virtual void eval_funct(const std::string& func, int nArg);

  /**  Dumped the statement to be evaluated as pased and stored in 
    *  reverse polish notation.
    */
  std::ostream& dump_rpn(std::ostream& out) const;

  int getDebug(void)const;

  /**  Set a debug printout level. Debugging information is printed to 
    *  the cerr stream depending on the requested level of detail.
    */
  void setDebug(int lvl);

  /** return a refernce to the evaluation stack.
   */
  const eval_stack& refStack(void) const;
  eval_stack& refStack(void);

private:
  /**  Internal recursive parsing method.
   */
   int recursive_parse(std::istream& in);

private:
  typedef std::vector<eval_stack::OpsEnum> op_stack;
  typedef std::vector<std::string> arg_stack;

private:
  int mOpLevel[eval_stack::kNOps];
  int mLevel;
  int mDebug;
  eval_stack& mEvalStack;
  op_stack mOpList;
  arg_stack mArgList;
  const char* mOpStr[eval_stack::kNOps];
};

inline int
Engine::getDebug(void) const {
    return mDebug;
}

inline eval_stack& 
Engine::refStack(void) {
    return mEvalStack;
}

inline const eval_stack& 
Engine::refStack(void) const {
    return mEvalStack;
}

#endif // !defined(ENGINE_HH)
