/*---------------------------------------------------------------------------*/
/*                                                                           */
/* Module Name:  TLGChannelBox						     */
/*                                                                           */
/* Module Description:  ROOT GUI widget for handling LIGO channel names      */
/*                                                                           */
/*---------------------------------------------------------------------------*/

#include "ligogui/TLGChannelBox.hh"
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <TVirtualX.h>
#include <TGClient.h>


namespace ligogui {
   using namespace std;


//--- Some utility functions ---------------------------------------------------
   static Int_t FontHeight(FontStruct_t f)
   {
      int max_ascent, max_descent;
      gVirtualX->GetFontProperties(f, max_ascent, max_descent);
      return max_ascent + max_descent;
   }

   static Int_t FontAscent(FontStruct_t f)
   {
      int max_ascent, max_descent;
      gVirtualX->GetFontProperties(f, max_ascent, max_descent);
      return max_ascent;
   }

   static Int_t FontTextWidth(FontStruct_t f, const char *c)
   {
      return gVirtualX->TextWidth(f, c, strlen(c));
   }



//////////////////////////////////////////////////////////////////////////
//                                                                      //
// ChannelTree                                                          //
//                                                                      //
// A list of channel names                                              //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

   ChannelTree::ChannelTree (const ChannelEntry* chns, UInt_t chnnum,
                     Int_t type)
   : fChnClient (0), fChannels (0), fChnNum (0), fChannelTreeType (type)
   {
      SetChannels (chns, chnnum);
   }


   ChannelTree::ChannelTree (const char* const* chnnames, UInt_t chnnum,
                     Int_t type)
   : fChnClient (0), fChannels (0), fChnNum (0), fChannelTreeType (type)
   {
      SetChannels (chnnames, chnnum);
   }


   ChannelTree::ChannelTree (const char* chnnames, Int_t type)
   : fChnClient (0), fChannels (0), fChnNum (0), fChannelTreeType (type)
   {
      if (chnnames != 0) {
         SetChannels (chnnames);
      }
   }


   ChannelTree::~ChannelTree ()
   {
      if (fChannels != 0) {
         delete [] fChannels;
         fChannels = 0;
         fChnNum = 0;
      }
   }


   Bool_t ChannelTree::SetChannels (const ChannelEntry* chns, 
                     UInt_t chnnum)
   {
      // allocate memory
      if (chnnum < 0) {
         return kFALSE;
      }
      if (fChannels != 0) {
         delete [] fChannels;
         fChannels = 0;
         fChnNum = 0;
      }
      if ((chnnum == 0) || (chns == 0)) {
         return kTRUE;
      }
      fChannels = new ChannelEntry [chnnum];
      if (fChannels == 0) {
         return kFALSE;
      }
      fChnNum = chnnum;
      for (UInt_t i = 0; i < fChnNum; ++i) {
         fChannels[i] = chns[i];
      //memcpy (fChannels, chns, fChnNum * sizeof (ChannelEntry));
      }
      return kTRUE;
   }


   Bool_t ChannelTree::SetChannels (const char* const* chnnames, 
                     UInt_t chnnum)
   {
      // allocate memory
      if (chnnum < 0) {
         return kFALSE;
      }
      if (fChannels != 0) {
         delete [] fChannels;
         fChannels = 0;
         fChnNum = 0;
      }
      if ((chnnum == 0) || (chnnames == 0)) {
         return kTRUE;
      }
      fChannels = new ChannelEntry [chnnum];
      if (fChannels == 0) {
         return kFALSE;
      }
      fChnNum = chnnum;
   
      // assign channels
      for (UInt_t i = 0; i < chnnum; i++) {
         fChannels[i].SetName (chnnames[i]);
         fChannels[i].SetUDN ("");
         fChannels[i].SetRate (0);
      }
      return kTRUE;
   }


   Bool_t ChannelTree::SetChannels (const char* chnnames)
   {
      return MakeChannelList (chnnames, fChannels, fChnNum);
   }


   Bool_t ChannelTree::MakeChannelList (const char* chnnames,
                     ChannelEntry*& chns, UInt_t& chnnum)
   {
      // allocate memory
      if (chns != 0) {
         delete [] chns;
         chns = 0;
         chnnum = 0;
      }
      if (chnnames == 0) {
         return kTRUE;
      }
      UInt_t size = 1;
      int i = 0;
      while  (chnnames[i] != 0) {
         if (isspace (chnnames[i])) {
            size++;
         }
         i++;
      }
      chns = new ChannelEntry [size];
      if (chns == 0) {
         return kFALSE;
      }
      // assign channel names
      char* p = new char [strlen (chnnames) + 1];
      if (p == 0) {
         return kFALSE;
      }
      strcpy (p, chnnames);
      char* lasts;
      char* tok = strtok_r (p, " \t\n\f\r\v", &lasts);
      while ((tok != 0) && (chnnum < size)) {
         chns[chnnum].SetName (tok);
         chns[chnnum].SetUDN ("");
         chns[chnnum].SetRate (0);
         tok = strtok_r (0, " \t\n\f\r\v", &lasts);
         // check if following token represents a UDN
         if (tok && (*tok == '@')) {
            chns[chnnum].SetUDN (tok + 1);
            tok = strtok_r (0, " \t\n\f\r\v", &lasts);
         }
         // check if following token represents sampling rate?
         bool num = true;
         for (char* c = tok; c && *c && num; c++) {
            num = isdigit (*c) || (*c == '.');
         }
         if (tok && *tok && num) {
            chns[chnnum].SetRate (atof (tok));
            tok = strtok_r (0, " \t\n\f\r\v", &lasts);
         }
         chnnum++;
      }
   
      delete [] p;
      return kTRUE;
   }


   Bool_t ChannelTree::GetIfoSub (const char* chn, 
                     char* ifo, char* sub, char* rest)
   {
      strcpy (ifo, "");
      strcpy (sub, "");
      char work[256];
      strncpy (work, chn, 255);
      work[255] = 0;
      char* p = strchr (work, ':');
      if (p == 0) {
         return kFALSE;
      }
      *p = 0;
      strncpy (ifo, work, 15);
      ifo[15] = 0;
      // vacuum channels!
      char* w = strchr (ifo, '-');
      if (w) {
         strncpy (sub, w + 1, 15);
         sub[15] = 0;
         *w = 0;
         strncpy (rest, p + 1, 255);
         rest[255] = 0;
         return  (strlen (ifo) == 3) && (strlen (sub) == 2) ? kTRUE : kFALSE;
      }
      // normal channels
      else {
         w = p + 1;
         p = strchr (w, '-');
         if (p == 0) {
            return kFALSE;
         }
         *p = 0;
         strncpy (sub, w, 15);
         sub[15] = 0;
         strncpy (rest, p + 1, 255);
         rest[255] = 0;
         return (strlen (ifo) == 2) && (strlen (sub) == 3) ? kTRUE : kFALSE;
      }
   }


   Bool_t ChannelTree::GetIfoSubLoc (const char* chn, 
                     char* ifo, char* sub, char* loc, char* rest)
   {
      strcpy (ifo, "");
      strcpy (sub, "");
      char work[256];
      strncpy (work, chn, 255);
      work[255] = 0;
      char* p = strchr (work, ':');
      if (p == 0) {
         return kFALSE;
      }
      *p = 0;
      strncpy (ifo, work, 15);
      ifo[15] = 0;
      // vacuum channels!
      char* w = strchr (ifo, '-');
      if (w) {
         strncpy (sub, w + 1, 15);
         sub[15] = 0;
         *w = 0;
         w = p + 1;
         p = strchr (w, '_');
         if (p == 0) {
            strcpy (rest, "*");
         }
         else {
            *p = 0;
            strncpy (rest, p + 1, 255);
            rest[255] = 0;
         }
         strncpy (loc, w, 255);
         loc[255] = 0;
         return  (strlen (ifo) == 3) && (strlen (sub) == 2) ? kTRUE : kFALSE;
      }
      // normal channels
      else {
         w = p + 1;
         p = strchr (w, '-');
         if (p == 0) {
            return kFALSE;
         }
         *p = 0;
         strncpy (sub, w, 15);
         sub[15] = 0;
         w = p + 1;
         p = strchr (w, '_');
         if (p == 0) {
            strcpy (rest, "*");
         }
         else {
            *p = 0;
            strncpy (rest, p + 1, 255);
            rest[255] = 0;
         }
         strncpy (loc, w, 255);	
         loc[255] = 0;
         return (strlen (ifo) == 2) && (strlen (sub) == 3) ? kTRUE : kFALSE;
      }
   }



extern "C" 
   int ChannelTree_channelcmp (const void* s1, const void* s2) {
      return (strcasecmp (((ChannelEntry*)s1)->Name(), 
                         ((ChannelEntry*)s2)->Name()));
   }

extern "C" 
   int ChannelTree_channelcmprate (const void* s1, const void* s2) {
      const ChannelEntry* c1 = (ChannelEntry*)s1;
      const ChannelEntry* c2 = (ChannelEntry*)s2;
      if (((c1->Rate() > 0) && (c1->Rate() <= 16)) ^
         ((c2->Rate() > 0) && (c2->Rate() <= 16))) {
         return ((c1->Rate() > 0) && (c1->Rate() <= 16)) ? 1 : -1;
      }
      else {
         return strcasecmp (c1->Name(), c2->Name());
      }
   }

   static TString displayedName (const char* rest, Int_t type, 
                     const ChannelEntry* chn)
   {
      TString entryname = rest;
      Bool_t openbra = kFALSE;
      if ((type & kChannelTreeShowRate) && (chn->Rate() > 0)) {
         char buf[256];
         sprintf (buf, " (%g", chn->Rate());
         entryname += buf;
         openbra = kTRUE;
      }
      if ((type & kChannelTreeShowSource) && (strlen (chn->UDN()) > 0)) {
         if (!openbra) entryname += "(";
         openbra = kTRUE;
         entryname += " @";
         entryname += chn->UDN();
      }
      if (openbra) entryname += ")";
      return entryname;
   }

   Bool_t ChannelTree::BuildChannelTree ()
   {
      DeleteChannels();
      if (fChannels == 0) {
         return kTRUE;
      }
   
      // sort first
      if (fChannelTreeType & kChannelTreeSeparateSlow) {
         ::qsort (fChannels, fChnNum, sizeof (ChannelEntry), 
                 ChannelTree_channelcmprate);
      }
      else {
         ::qsort (fChannels, fChnNum, sizeof (ChannelEntry), 
                 ChannelTree_channelcmp);
      }
   
      // go through list and add to tree
      char prevIfo[64] = {0};
      char ifo[64] = {0};
      TLGLBTreeEntry* prevIfoItems = 0;
      char prevSub[64] = {0};
      char sub[64] = {0};
      TLGLBTreeEntry* prevSubItems = 0;
      char prevLoc[256] = {0};
      char loc[256] = {0};
      TLGLBTreeEntry* prevLocItems = 0;
      char rest[256] = {0};
      int* oddItems = new int [fChnNum];
      int oddNum = 0;
      if (oddItems == 0) {
         return kFALSE;
      }
   
      for (UInt_t i = 0; i < fChnNum; i++) {
         bool succ = false;
         if (fChannelTreeType & kChannelTreeLevel3) {
            succ = GetIfoSubLoc (fChannels[i].Name(), ifo, sub, loc, rest);
         }
         else {
            succ = GetIfoSub (fChannels[i].Name(), ifo, sub, rest);
         }
         if (succ) {
            if (fChannelTreeType & kChannelTreeSeparateSlow) {
               if ((fChannels[i].Rate() > 0) && 
                  (fChannels[i].Rate() <= 16)) {
                  strcpy (ifo + strlen (ifo), " (slow)");
               }
            }
            // create site/ifo branch if necessary
            if ((strcasecmp (prevIfo, ifo) != 0) ||
               (prevIfoItems == 0)) {
               prevIfoItems = AddChannel (0, ifo);
               strcpy (prevIfo, ifo);
               prevSubItems = 0;
               strcpy (prevSub, "");
            }
            if (prevIfoItems == 0) {
               delete [] oddItems;
               return kFALSE;
            }
            // create subsystem branch if necessary
            if ((strcasecmp (prevSub, sub) != 0) ||
               (prevSubItems == 0)) {
               prevSubItems = AddChannel (prevIfoItems, sub);
               strcpy (prevSub, sub);
            }
            if (prevSubItems == 0) {
               delete [] oddItems;
               return kFALSE;
            }
            // create location branch if necessary
            if (fChannelTreeType & kChannelTreeLevel3) {
               if ((strcasecmp (prevLoc, loc) != 0) ||
                  (prevLocItems == 0)) {
                  prevLocItems = AddChannel (prevSubItems, loc);
                  strcpy (prevLoc, loc);
               }
               if (prevLocItems == 0) {
                  delete [] oddItems;
                  return kFALSE;
               }
            }
            // create channel entry
            TLGLBTreeEntry* pItems = (fChannelTreeType & kChannelTreeLevel3) ?
               prevLocItems : prevSubItems;
            TString entryname = 
               displayedName (rest, fChannelTreeType, fChannels + i);
            if (fChnClient == 0) {
               AddChannel (pItems, entryname, fChannels[i].Name(), i);
            }
            else {
               AddChannel (pItems, entryname, fChannels[i].Name(), i,
                          fChnClient->GetPicture("arrow_left.xpm"), 
                          fChnClient->GetPicture("arrow_left.xpm"));
            }
         }
         // oddball channel names: mark them and add at end
         else {
            oddItems[oddNum] = i;
            oddNum++;
         }
      }
   
      // check odd channel names
      if (oddNum > 0) {
         prevIfoItems = AddChannel (0, "others");
         if (prevIfoItems == 0) {
            delete [] oddItems;
            return kFALSE;
         }
         for (int i = 0; i < oddNum; i++) {
            TString entryname = 
               displayedName (fChannels[oddItems[i]].Name(), 
                             fChannelTreeType, fChannels + oddItems[i]);
            if (fChnClient == 0) {
               AddChannel (prevIfoItems, entryname,
                          fChannels[oddItems[i]].Name(), oddItems[i]);
            }
            else {
               AddChannel (prevIfoItems, entryname, 
                          fChannels[oddItems[i]].Name(), oddItems[i], 
                          fChnClient->GetPicture("arrow_left.xpm"), 
                          fChnClient->GetPicture("arrow_left.xpm"));
            }
         }
      }
   
      delete [] oddItems;
      return kTRUE;
   }


   Int_t ChannelTree::GetChannelId (const char* channelname) const 
   {
      for (UInt_t i = 0; i < fChnNum; i++) {
         if (strcasecmp (fChannels[i].Name(), channelname) == 0) {
            return i;
         }
      }
      return -1;
   }


   const char* ChannelTree::GetChannelName (Int_t id) const
   {
      if ((id < 0) || ((UInt_t)id >= fChnNum)) {
         return 0;
      }
      else {
         return fChannels[id].Name();
      }
   }



//////////////////////////////////////////////////////////////////////////
//                                                                      //
// TLGChannelListbox                                                    //
//                                                                      //
// A tree list box to deal with channel names                           //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

   TLGChannelListbox::TLGChannelListbox (TGWindow* p, Int_t id, 
                     ChannelEntry* chnnames, UInt_t chnnum, 
                     Int_t type, UInt_t option, ULong_t back)
   :  TLGLBTree (p, id, option, back), 
   ChannelTree (chnnames, chnnum, type)
   {
      fChnClient = fClient;
      BuildChannelTree();
   }


   TLGChannelListbox::TLGChannelListbox (TGWindow* p, Int_t id, 
                     const char* chnnames, Int_t type, 
                     UInt_t option, ULong_t back)
   : TLGLBTree (p, id, option, back), ChannelTree (chnnames, type)
   {
      fChnClient = fClient;
      BuildChannelTree();
   }


   TLGChannelListbox::~TLGChannelListbox ()
   {
   }



//////////////////////////////////////////////////////////////////////////
//                                                                      //
// TLGChannelCombobox                                                   //
//                                                                      //
// A tree combobox to deal with channel names                           //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

   TLGChannelCombobox::TLGChannelCombobox (TGWindow* p, Int_t id, 
                     ChannelEntry* chnnames, UInt_t chnnum,
                     Int_t type, Bool_t editable, 
                     UInt_t option, ULong_t back)
   : TLGComboTree (p, id, editable, option), 
   ChannelTree (chnnames, chnnum, type)
   {
      fChnClient = fClient;
      BuildChannelTree();
   }


   TLGChannelCombobox::TLGChannelCombobox (TGWindow* p, Int_t id, 
                     const char* chnnames, Int_t type, 
                     Bool_t editable, UInt_t option, ULong_t back)
   : TLGComboTree (p, id, editable, option), ChannelTree (chnnames, type)
   {
      fChnClient = fClient;
      BuildChannelTree();
   }


   TLGChannelCombobox::~TLGChannelCombobox ()
   {
   }


}
