#ifndef ISMBUF_HH_
#define ISMBUF_HH_
//
//   Shared memory stream input buffer class
//
#include "PConfig.h"
// #define ISMBUF_TRACE

#if defined(ISMBUF_TRACE)
  #include <string>
  #include <iostream>
#endif
#if defined(__GNU_STDC_OLD)
   #include <streambuf.h>
#else
   #include <iostream>
   #include <streambuf>
#endif

class LSMP_CON;

/**  This class interfaces the shared memory buffer manager to a standard
  *  data stream. iSMbuf is based on the standard stream buffer (streambuf)
  *  class and implements the virtual buffering methods of that class. In
  *  addition to the standard methods, the iSMbuf class contains waitBuf
  *  and setBCount methods to wait for asynchronous data and to set a 
  *  reserved buffer count, respectively. The buffer functions ignore the
  *  arrival of signals, except in the waitBuf method.
  *  @memo Shared memory stream input buffer class.
  */
class iSMbuf : public std::streambuf {
  public:
    /**  Default constructor.
      *  The default constructor builds an unopened shared memory stream 
      *  buffer.
      */
    iSMbuf();

    /**  Construct an opened shared memory stream buffer.
      *  A shared memory stream buffer is created and attached (if possible) 
      *  to the specified shared memory partition. The partition to be 
      *  attached is specified by 'partition'. The mode flag is kept for
      *  compatibility with the file stream and must be set to ios::in.
      */
#ifdef __GNU_STDC_OLD
    iSMbuf(const char *partition, std::ios::open_mode mode=std::ios::in);
#else
    iSMbuf(const char *partition, std::ios::openmode mode=std::ios::in);
#endif

    /**  Shared memory stream buffer destructor.
      */
    ~iSMbuf();

    /**  Attach a shared memory partition to a stream buffer.
      *  The named partition is attached to the stream buffer.
      */
#ifdef __GNU_STDC_OLD
    iSMbuf* open(const char *partition, std::ios::open_mode mode=std::ios::in);
#else
    iSMbuf* open(const char *partition, std::ios::openmode mode=std::ios::in);
#endif

    /**  This parameter controls what fraction of the data read into 
      *  the shared memory will be copied to the data stream and what
      *  part of the shared memory resources can be held allocated the 
      *  stream. 'nbuf' == 0 indicates data is taken as requested.
      *  'nbuf' == -1 indicates that all data written to the shared 
      *  memory must be copied to the data stream. 'nbuf'==N indicates
      *  that up to N buffers (frames) of data should be held until the 
      *  stream is ready for them.
      *  @memo Set the number of segments to be reserved by this consumer.
      */
    void setBCount(int nbuf);

    /**  Test whether data are available. If data are available, waitBuf
      *  returns true. If data are not immediately available and the nowait
      *  argument is true, waitData returns false. Otherwise waitBuf waits
      *  for data to arrive or the semaphore wait to be interrupted by a 
      *  signal.
      *  @memo Wait for data.
      *  @param nowait If true, waitBuf returns immediately.
      *  @return true if data available.
      */
    bool waitBuf(bool nowait=false);

    /**  Release a shared memory buffer.
      *  @memo Release shared memory buffer.
      */
    void relse(void);

    /**  Standard streambuf allocation function.
      */
    int doallocate(void);

    /**  Standard streambuf overflow function.
      */
    int overflow(int c) {return EOF;}

    /**  Standard streambuf read function.
      */
    int underflow(void);

#ifdef __GNU_STDC_OLD
    /**  Standard streambuf seek function.
      */
    std::streampos seekoff(std::streamoff off, std::ios::seekdir dir,
			   int mode);
#else
    typedef std::streambuf::off_type off_type;
    typedef std::streambuf::pos_type pos_type;

    /**  Standard streambuf seek function.
      */
    pos_type seekoff(off_type off, std::ios_base::seekdir dir,
		     std::ios_base::openmode mode);

    pos_type seekpos(pos_type off, std::ios_base::openmode mode);
#endif

    /**  Standard streambuf synchronization function.
      */
    int sync(){return 0;}

  private:
    void setptrs(int ix=0);
#ifndef __GNUG__
  void setb(char* b, char*eb, int a=0) {}
#endif


  private:
    ///                  Pointer to the shared memory consumer object.
    LSMP_CON*   mConsumer;
    ///                  Pointer to the current shared memory segment.
    const char* mBuffer;
    ///                  Data length in the current shared memory segment.
    int         mLength;
#ifdef ISMBUF_TRACE
  public:
    void pTrace(void) const {mTrace.pTrace();}
  private:
    class trace {
    #define maxTrace 8
    public:
        trace() : nTrace(0) {}
        void tracepoint(const std::string& funct, int arg1=0, int arg2=0);
        void pTrace(void) const;
        void clear(void);
    private:
        int nTrace;
        std::string mFuncName[maxTrace];
        int mArg1[maxTrace];
        int mArg2[maxTrace];
    } mTrace;
#endif
};

//======================================  Inline implementation of seekoff
#ifndef __GNU_STDC_OLD
inline iSMbuf::pos_type
iSMbuf::seekpos(pos_type off, std::ios_base::openmode mode) {
    return seekoff(off_type(off), std::ios::beg, mode);
}
#endif

//======================================  Trace point function
#ifdef ISMBUF_TRACE
inline void 
iSMbuf::trace::tracepoint(const std::string& funct, int arg1, int arg2) {
    int inx = nTrace % maxTrace;
    mFuncName[inx] = funct;
    mArg1[inx] = arg1;
    mArg2[inx] = arg2;
    nTrace++;
}

inline void 
iSMbuf::trace::clear(void) {
    nTrace = 0;
}
#endif

#endif  //  ISMBUF_HH_
