#ifndef SEGGENER_HH
#define SEGGENER_HH

#include "DatEnv.hh"
#include "OperStateCondList.hh"
#include "TrigClient.hh"
#include "Time.hh"
#include <list>
#include <string>

namespace trig {
  class Segment;
}

//
//  Segment condition definition class.
//
//  The segment condition class contains all the information needed to 
//  define a segment.
class SegCond {
public:
  typedef unsigned int mask_t;

public:
  //  Null Constructor
  SegCond(void);

  //  Constructor for segments that follow a signal condition.
  SegCond(const std::string& group, int v, const std::string& on);

  //  Constructor for segments turned on by one condition and off by another.
  SegCond(const std::string& group, int v, const std::string& on, 
	  const std::string& off);

  //  Destructor.
  ~SegCond(void);

  //  Evaluate the condition. Evaluate sets mInSegment if the condition 
  //  is asserted and mSegComplete when mInSegment transitions from true
  //  to false.
  bool evaluate(OperStateCondList& l, const Time& t, Interval dT);

  //  testPrevious returns the segment state before the last completed
  //               evaluate.
  bool testPrevious(void) const;

  //  testSegment returns true if mSegComplete is set, i.e. after the 
  //  segment starts.
  bool testSegment(void) const;

  //  True if the segment condition is currently asserted
  bool testInSegment(void) const;

  //  Return the 
  const char* getGroup(void) const;


  Time getEndTime(void) const;
  Interval getLimit(void) const;
  Interval getLiveTime(void) const;
  int getNSegment(void) const;
  const char* getSegFile(void) const;
  const char* getStartCond(void) const;
  Time getStartTime(void) const;
  Time getSegBegin(void) const;
  Time getSegEnd(void) const;
  const char* getStopCond(void) const;
  int getVersion(void) const;
  trig::Segment makeSegment(void);
  void reset(void);
  std::string& refGroup(void);
  const std::string& refIfo(void) const;
  const std::string& refStartCond(void) const;
  void setComment(const std::string& comm);
  void setIfo(const std::string& ifo);
  void setLimit(Interval dT);
  void setSegFile(const std::string& file);
  void update(void);

private:
  //------------------------------------  Parameters
  std::string mGroup;
  int         mVersion;
  Interval    mLimit;
  enum {
    kUnknown,
    kStartCond,  // Segment from "on" until "off"
    kOnCond      // Segment while "on"
  }           mType;
  std::string mOnCond;
  std::string mOffCond;
  std::string mSegFile;
  std::string mComment;
  std::string mIfo;

  //------------------------------------  State
  bool mPrevious;
  bool mInSegment;

  //  Time of last transition from Off -> On.
  Time mStart;

  //  Time of last transition from On -> Off.
  Time mStop;

  //  End time of last evaluated segment
  Time mCurrent;

  //------------------------------------  Statistics
  Interval    mLiveTime;
  int         mNSegment;
};

//======================================  Inline functions
inline const char* 
SegCond::getGroup(void) const {
    return mGroup.c_str();
}

inline Time 
SegCond::getEndTime(void) const {
    return mStop;
}

inline Interval
SegCond::getLimit(void) const {
    return mLimit;
}

inline Interval
SegCond::getLiveTime(void) const {
    if (mInSegment) return mLiveTime + (mCurrent - mStart);
    return mLiveTime;
}

inline int
SegCond::getNSegment(void) const {
    return mNSegment;
}

inline Time
SegCond::getSegBegin(void) const {
    if (mPrevious) return mStart;
    return mStop;
}

inline Time
SegCond::getSegEnd(void) const {
    if (mPrevious) return mStop;
    return mStart;
}

inline const char*
SegCond::getSegFile(void) const {
    return mSegFile.c_str();
}

inline Time 
SegCond::getStartTime(void) const {
    return mStart;
}

inline const char*
SegCond::getStartCond(void) const {
    return mOnCond.c_str();
}

inline const char*
SegCond::getStopCond(void) const {
    if (mType == kOnCond) return "-";
    return mOffCond.c_str();
}

inline int 
SegCond::getVersion(void) const {
    return mVersion;
}

inline const std::string& 
SegCond::refIfo(void) const {
    return mIfo;
}

inline const std::string& 
SegCond::refStartCond(void) const {
    return mOnCond;
}

inline std::string& 
SegCond::refGroup(void) {
    return mGroup;
}

inline bool 
SegCond::testInSegment(void) const {
    return mInSegment;
}

/**  @name SegGener
  *  SegGener generates entries to the segment meta-database table based
  *  on conditions evaluated by the OperStateCond (OSC) class. The user
  *  specifies one or more segment types to be evaluated in a configuration
  *  file. The segment may be asserted either at all times that a condition
  *  is true, or from the onset of a start condition to the onset of a stop
  *  condition. A segment table row is generated each time a stretch of
  *  data is observed where the specified condition is true. SegGener will
  *  optionally split the segments into spans of a given maximum length.
  *  This allows the segment to be generated before the condition returns
  *  to false in cases where the condition remains asserted over long 
  *  periods of time.
  *
  *  {\bf {\Large Running SegGener} }
  *
  *  The SegGener command line is as follows:
  *
  *  \begin{verbatim}
      SegGener -conf <config-file> -osc <osc-defs> [+seg] [-htfile <report>] 
     \end{verbatim}
  *
  *  where:
  *  \begin{tabular}{ll}
  *      #<config-file># & Configuration file name \\
  *      #<osc-defs>#    & OSC definition file name \\
  *      #<report>#      & Html report file name \\
  *      #+seg#          & Write segment to LDAS meta-database (via TrigMgr).
  *  \end{tabular}
  *
  *  {\bf {\Large The SegGener configuration file} }
  *
  *  Each line of the SegGener configuration file defines one segment to 
  *  be evaluated and (optionally) written. Each configuration file line
  *  has the following format:
  *
  *  \begin{verbatim}
      <segment-name> <segment-version> <start-cond> [<stop-cond>] \
                     [-limit <max-time>] [-file <seg-file>]
     \end{verbatim}
  *
  *  where #<segment-name># is the segment name, #<segment-version># is the
  *  segment version number. The segment condition is defined by 
  *  #<start-cond># and #<stop-cond>#. If #<stop-cond># is defined, the 
  *  Segment is asserted from the onset of #<start-cond># to the onset
  *  of #<stop-cond>#. If #<stop-cond># is not defined, the segment lasts
  *  as long as #<start-cond># is asserted. The #-limit# argument specifies
  *  the maximum time (#<max-time>#, in seconds) for a single segment table 
  *  entry. If a segment condition lasts longer than the specified time, 
  *  SegGener will write multiple segment table rows after the segment rows
  *  with durations less than or equal to the specified maximum. The #-file#
  *  argument specifies a file name (default directory is #$DMTHTMLOUT#) 
  *  to which one line will be appended for each segment.
  *
  *  {\bf {\Large SegGener outputs} }
  *
  *  {\bf Web readable report file}
  *
  *  SegGener produces a web readable report file with a table containing 
  *  the current segment definitions and a statistical summary of the
  *  meta-database rows produced by each segment type. The segment report
  *  is in #html# format and is written to #$DMTHTML/Segment_List.html#.
  *  The segment statistics table has the following entries:
  *  \begin{tabular}{ll}
  *  Column & Contents \\ \hline
  *  Segment Group & Name of the segment group to be written \\
  *  In Segment & Indicates whether the segment condition was true at the
  *               time the report was written \\
  *  Start Condition & OSC condition marking the start or duration 
  *               of the segment. \\
  *  Stop Condition & OSC condition marking the end of the segment (or '-'
  *                   if the segment is defined by a single condition) \\
  *  Live &       Fraction of the time that the segment is asserted (in \%) \\
  *  Segments Generated & Number of segments that have been generated. \\
  *  Time Limit & Maximum time of a single segment entry. \\
  *  \end{tabular}
  *
  *  {\bf Segment metadata}
  *
  *  Each detected segment is sent to the Trigger Manager to be written to 
  *  the LDAS meta-database if the #+seg# command line argument is specified.
  *
  *  {\bf Segment list file}
  *
  *  In addition to the metadata, the segment may be optionally written to an
  *  ascii segment list file. Each line in the list file contains the line 
  *  number, sstart time, stop time and duration for a single segment or 
  *  segment fragment. The line is appended to the file at the same time the
  *  segment info is sent.
  * 
  *  @memo Generate segment database entries from OSC conditions.
  *  @author John Zweizig
  *  @version 1.1; Last modified June 10, 2002
  */
//{
//}


class SegGener : public DatEnv, TrigClient {
public:
  SegGener(int argc, const char* argv[]);
  ~SegGener(void);
  void ProcessData(void);
  bool readConfig(const char* file);
  void writeTable(void) const;

private:
  bool OSCexists(const std::string& s) const;

private:
  typedef std::list<SegCond> SegList;
  typedef SegList::const_iterator const_iter;
  typedef SegList::iterator       list_iter;

private:
  std::string mConfig;
  std::string mOSCfile;

  SegList mList;
  OperStateCondList mOSC;
  bool    mRecord;
  Time    mFirst;
  Time    mLast;
  std::string mHtFile;
};

inline bool
SegGener::OSCexists(const std::string& s) const {
    return (mOSC.find(s) != mOSC.end());
}

#endif  //  SEGGENER_HH
