#include "magGlitch.hh"

#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cmath>
#include <iomanip>
#include <string>
#include <vector>
#include "DVector.hh"
#include "Complex.hh"

using namespace std;

extern void ApplyFilter(int,const int, const float*,float*,
			const double*,const double*, int);
extern void GetFilterCoef(const int,const int,const float,const float, 
			  double*,double*,const int);


EXECDAT(magGlitch);


// constructor
magGlitch::magGlitch(int argc, const char *argv[]) :  
  DatEnv(argc,argv),fDAQInterval(0),fLogInterval(3600),fDecayTime(10),
  fLogName("magGlitch.log"),fDataName("magGlitch.dat"),fTriggerON(0) {

  Reset();

  string ConfigName("magGlitch.conf");
  const char *parsname = getenv("DMTPARS");
  const char *sitename = getenv("LIGOSITE");
  if (parsname && sitename) 
    ConfigName= (string)parsname + "/magGlitch_" + (string)sitename + ".conf";
  else if (sitename)
    ConfigName= "magGlitch_" + (string)sitename + ".conf";

  //set config file name
  if (argc>1){
    ConfigName = argv[1];
  }
  
  //set other parameters
  MonServer::setDebug(Debug());
  MonServer::setServerName(argv[0]);

  //open config file
  ifstream ConfigFile;
  ConfigFile.open(ConfigName.c_str(),ios::in);
  if (!ConfigFile.is_open()){
    cerr<<"Configuration File Not Found."<<endl;
    exit(-1);
  }

  // read config file
  SetConfig(ConfigFile);
 
  ConfigFile.close();

  // allocate buffer
  fData        = new float[16384 * 2];
  fFilterData  = new float[16384 * kOffsetFactor];
  
  //open log file
  fLogFile.open(fLogName.c_str(),ios::out|ios::app);
  if (!fLogFile.is_open()){
    cerr<<fLogName<<" Log File Error"<<endl;
    exit(-1);
  }
	
  //open data file
  fDataFile.open(fDataName.c_str(),ios::out|ios::app);
  if (!fDataFile.is_open()){
    cerr<<fDataName<<" Data File Error"<<endl;
    exit(-1);
  } 


  // output config info to Log file
  fLogFile << "====================================" << endl
	   << "Data File      : " << fDataName << endl
	   << "DAQ Interval   : " << fDAQInterval<< endl
	   << "Decay Time     : " << fDecayTime << endl
	   << "Trigger ON/OFF : " << fTriggerON << endl
	   << "Log File       : " << fLogName << endl
	   << "Log Interval   : " << fLogInterval << endl
	   << "# of Channels  : " << fCHStatus.size() << endl;
  for (vector<ChannelStatus>::iterator i = fCHStatus.begin();
       i != fCHStatus.end(); ++i) {

    fLogFile<<i->fName<<" "
	    <<i->fSampleFreq<<" "
	    <<i->fThresholdLow<<" "
	    <<i->fThresholdHi<<" "
	    <<i->fFilterMode<<" "
	    <<i->fFilterOrder<<" "
	    <<i->fCutOffFreqLow<<" "
	    <<i->fCutOffFreqHi<<" "
	    <<i->fTrigComment
	    <<endl;

  }

}


// destructor
magGlitch::~magGlitch(void)
{
  fLogFile<<"Monitor End = "
	  <<fCHStatus.begin()->fTSnew->getEndTime().getS()<<endl
	  <<"\t"<<double(fCount)/fLogInterval<<" intervals of "
	  <<fLogInterval<<" seconds"<<endl
	  <<"Final Aggregated Histograms:"<<endl
	  <<"===================================="<<endl;
  for (vector<ChannelStatus>::iterator i = fCHStatus.begin();
       i != fCHStatus.end(); ++i) i->fHistTot->Dump(fLogFile);

  if (fData) delete[] fData;
  if (fFilterData) delete[] fFilterData;
  fLogFile.close();
  fDataFile.close();
}


void magGlitch::Attention(void) {
  if (Debug()) cout << "magGlitch Attention."<<endl;
  MonServer::Attention();
  for (vector<ChannelStatus>::iterator i=fCHStatus.begin();
       i != fCHStatus.end(); ++i) i->fHistTot->Dump(fLogFile);
}

void magGlitch::SetConfig(ifstream& config)
{

  std::string line;
  std::vector<string> element;
  std::string::size_type pos, prev_pos;

  // read and set configurations
  while (!config.eof()) {

    // read one line from config file 
    getline(config,line);
    line += " ";

    // separate element in the line
    pos = prev_pos = 0;
    while( ( pos = line.find_first_of(" ",pos)) != string::npos ) {
      string word = line.substr( prev_pos, pos - prev_pos);
      if ( word.size() ) element.push_back(word.c_str());
      prev_pos = ++pos;
    }


    // set global and channel parameters
    ChannelStatus chstat;
    int index = kConfigTag;
    const char* datadir = getenv("DMTOUTPUT");
    const char* htmldir = getenv("DMTHTMLOUT");
    char* startdate=0;
    TimeStr(Now(),startdate,"_%y%m%d");
    for ( vector<string>::iterator itr = element.begin(); 
	  itr != element.end(); ++itr ) {
      if ( *element.begin() == "DataFilePath" ) {
	if (index == kConfigDataFilePath) {
	  if (datadir) fDataName = (string)datadir + "/";
	  fDataName += *itr;
	  string movedata= string("mv -f ")+fDataName+" "+fDataName+startdate;
	  system(movedata.c_str());
	}
	else index = kConfigDataFilePath;
      }
      else if ( *element.begin() == "DAQInterval" ) {
	if (index == kConfigDAQInterval) {
	  fDAQInterval = atoi(itr->c_str());
	}
  	else index = kConfigDAQInterval;
      }
      else if ( *element.begin() == "DecayTime" ) {
	if (index == kConfigDecayTime) {
	  fDecayTime = atoi(itr->c_str());
	}
	else index = kConfigDecayTime;
      }
      else if ( *element.begin() == "TriggerON" ) {
	if (index == kConfigTriggerON) {
	  fTriggerON = (bool) atoi(itr->c_str());
	}
	else index = kConfigTriggerON;
      }
      else if ( *element.begin() == "LogFilePath" ) {
	if (index == kConfigLogFilePath) {
	  if (htmldir) fLogName = (string)htmldir + "/";
	  fLogName += *itr;
	  string movelog = string("mv -f ")+fLogName+" "+fLogName+startdate;
	  system(movelog.c_str());
	}
	index = kConfigLogFilePath;
      }
      else if ( *element.begin() == "LogInterval" ) {
	if (index == kConfigLogInterval) {
	  fLogInterval = atoi(itr->c_str());
	}
	else index = kConfigLogInterval;
      }
      else if ( *element.begin() == "Channel" ) {
	if ( *itr == "End" ) fCHStatus.push_back(chstat);
	else {
	  switch (index) {
	  case kConfigTag:
	    index = kConfigChannelName;
	    break;
	  case kConfigChannelName:
	    chstat.fName = *itr;
	    index = kConfigSampleFreq;
	    break;
	  case kConfigSampleFreq:
	    chstat.fSampleFreq = atoi(itr->c_str());
	    index = kConfigThresholdLow;
	    break;
	  case kConfigThresholdLow:
	    chstat.fThresholdLow = atof(itr->c_str());
	    index = kConfigThresholdHi;
	    break;
	  case kConfigThresholdHi:
	    chstat.fThresholdHi = atof(itr->c_str());
	    index = kConfigFilterMode;
	    break;
	  case kConfigFilterMode:
	    if (strcasecmp(itr->c_str(),"LowPass") == 0) {
	      chstat.fFilterMode = ChannelStatus::kLowPass;
	      index = kConfigFilterOrder;
	    }
	    else if (strcasecmp(itr->c_str(),"HighPass") == 0) {
	      chstat.fFilterMode = ChannelStatus::kHighPass;
	      index = kConfigFilterOrder;
	    }
	    else if (strcasecmp(itr->c_str(),"BandPass") == 0) {
	      chstat.fFilterMode = ChannelStatus::kBandPass;
	      index = kConfigFilterOrder;
	    }
	    else { // NoFilter
	      chstat.fFilterMode = ChannelStatus::kNoFilter;
	      index = kConfigTrigComment;
	    }
	    break;
	  case kConfigFilterOrder:
	    chstat.fFilterOrder = atoi(itr->c_str());
	    index = kConfigCutOffFreqLow;
	    break;
	  case kConfigCutOffFreqLow:
	    chstat.fCutOffFreqLow = atof(itr->c_str());
	    if ( chstat.fFilterMode == ChannelStatus::kLowPass ||
		 chstat.fFilterMode == ChannelStatus::kHighPass) {
	      index = kConfigTrigComment;
	    }
	    else index = kConfigCutOffFreqHi;
	    break;
	  case kConfigCutOffFreqHi:
	    chstat.fCutOffFreqHi = atof(itr->c_str());
	    index = kConfigTrigComment;
	    break;
	  case kConfigTrigComment:
	    chstat.fTrigComment = *itr;
	    index = 999;
	    break;
	  }
	}
      }
    }
    element.clear();
  }

  fNChannel = fCHStatus.size();
  if (!fNChannel) { // kill if fNChannel or fFilterOrder <=0
    if (!fNChannel) cerr << "# of Channels <= 0" << endl;
    exit(-1);
  }

  // set decay constants for averaging sdev
  fDecay = exp(-log(2.0)/fDecayTime);

  // allocate time series and histograms and set channels to data access,
  // then set filter coefficients
  int j = 0;
  for (vector<ChannelStatus>::iterator i = fCHStatus.begin();
       i != fCHStatus.end(); ++i) {
    
    i->fTSnew = new TSeries(Time(0),Interval(0.0),i->fSampleFreq);
    i->fTSold = new TSeries(Time(0),Interval(0.0),i->fSampleFreq);
    i->fHistCur  = new Histogram1();
    i->ResetHist(Now());
    i->fHistTot  = new Histogram1(*(i->fHistCur));
    string temp= string(i->fHistTot->GetTitle()) + "- Accumulated";
    i->fHistTot->SetTitle(temp.c_str());
    temp = string(i->fHistCur->GetNLabel()) + " 0 intervals";
    i->fHistTot->SetNLabel(temp.c_str());

    getDacc().addChannel(i->fName.c_str(),1,&(i->fTSnew));
    if (Debug()) cout << "Serving histogram: " << i->fHistCur->GetTitle() 
		      << endl;
    serveData(i->fHistCur->GetTitle(), i->fHistCur, "1-D Histogram");
    if (Debug()) cout << "Serving histogram: " << i->fHistTot->GetTitle()
		      << endl;
    serveData(i->fHistTot->GetTitle(), i->fHistTot, "1-D Histogram");

    i->ResetFilterCoef();
    GetFilterCoef(i->fFilterOrder,i->fSampleFreq,
		  i->fCutOffFreqLow,i->fCutOffFreqHi,
		  i->fCoefA,i->fCoefB,i->fFilterMode);

    i->fInitCount = 1;
    j++;
  }


}



// main loop
void magGlitch::ProcessData(void)
{


  for (vector<ChannelStatus>::iterator ch = fCHStatus.begin();
       ch != fCHStatus.end(); ch++) {

    if (ch->IsTSContinuous()){ // check continuity

      // copy TSeries to data array
      ch->CopyTStoArray(fData);

      // calc average
      ch->fMean = GetMean(&(*ch));

      // remove offset and move last kOffsetFactor second out of 2-sec data
      // to the beginning of the same data array
      for (int j = 0; j < ch->fSampleFreq * kOffsetFactor ; j++){
	fData[j]=fData[ch->fSampleFreq * (kOffsetFactor - 1) + j] - ch->fMean;
      }

      if (ch->fEventStatus == ChannelStatus::kNone) {
      float sdevorg = GetSDev(ch->fSampleFreq);
      if (sdevorg) {
        ch->fSumSDev += sdevorg;
        ch->fSumSDevCount++;
      }
      }

      if (ch->fFilterMode == ChannelStatus::kNoFilter) {
	for (int j = 0; j < ch->fSampleFreq * kOffsetFactor; j++)
	  fFilterData[j] = fData[j];
      }
      else { // apply IIR filter (kOffsetFactor second of time series)
	ApplyFilter(ch->fFilterOrder,ch->fSampleFreq * kOffsetFactor,
		    fData,fFilterData,ch->fCoefA,ch->fCoefB,ch->fFilterMode);  
      }

      CheckEvent(&(*ch));
      ch->fSumMean += ch->fMean; // sum of mean
      ch->fCheckEventCount++; // increment # of CheckEvent executed normally
      if (ch->fInitCount) ch->fInitCount--;

    }
    else { // handle continuity error 
	
      if (!ch->fInitCount) {
	fLogFile<<"TimeSeries Continuity Error "
		<<ch->fName<<" "
		<<ch->fTSnew->getStartTime().getS()
		<<endl;
      }

    }

    // copy current second 1-sec data to first 1-sec for the next
    *(ch->fTSold) = *(ch->fTSnew);
  }

  fCount++; // increment # of ProcessData executed
  OutputLog();

  if (fDAQInterval && fCount >= fDAQInterval) { //check the end of run
    finish();
  }

}



// glitch status check
// data holds filtered time series
void magGlitch::CheckEvent(ChannelStatus* ch)
{
  float sdev = 0;
  int sdevcount = 0;
  float sdevorg = 0;
  int sdevorgcount = 0;
  
  // reject channel with sdev less than 1 count or saturated
  if (ch->fSDev > 1 &&
      (abs(ch->fMean) + 3 * ch->fSDev) < 32768 &&
      !ch->fInitCount) {

    int j = 0;
    while (j < ch->fSampleFreq) {

      Time currenttime = ch->fTSnew->getBinT(j);
      float dataF = fFilterData[ch->fSampleFreq * (kOffsetFactor - 1) + j];
      float data = fData[ch->fSampleFreq * (kOffsetFactor - 1) + j];

      CheckEvent(data,currenttime,ch);

      // sdev calc
      if (ch->fEventStatus == ChannelStatus::kNone || 
        ch->GetEventDuration() > kMaxSDevSuspend) { 
      sdev += dataF * dataF;
      sdevcount++;

      sdevorg += data * data;
      sdevorgcount++;
      }
      
      j++;

    }

    // sdev: filtered data
    if (!sdevcount) sdev = -1;
    else sdev = sqrt(sdev/sdevcount);

    // sdev: unfiltered data
    if (!sdevorgcount) sdevorg = -1;
    else sdevorg = sqrt(sdevorg/sdevorgcount);
    
  }
  else { // calc sdev: low sdev, saturation, initialization
    sdev = GetSDevF(ch->fSampleFreq);
    sdevorg = GetSDev(ch->fSampleFreq);
  }

  if ( sdev >= 0 ) UpdateSDevF(sdev,ch);

  if ( sdevorg >= 0 ) {
    ch->fSumSDev += sdevorg;
    ch->fSumSDevCount++;
  }

  return;

}


void magGlitch::CheckEvent(const float& data, const Time& currenttime, ChannelStatus* ch)
{
  //check glitch
  if ( abs(data) > ch->fThresholdLow * ch->fSDev){ // data above threshold

    if (ch->fEventStatus == ChannelStatus::kNone){ // event start

      ch->fStartTime   = currenttime;
      ch->fMaxTime     = currenttime;
      ch->fMaxData     = abs(data)/ch->fSDev;
      ch->fMaxSDev     = ch->fSDev;
      ch->fMaxMean     = ch->fMean;
      ch->fEventStatus = ChannelStatus::kEvent;
      ch->fEventNPoint = 1; 
      
    } 
    else{ // event continue

      ch->fEventNPoint++;
      ch->fEventStatus = ChannelStatus::kEvent;
      if (abs(data)/ch->fSDev > ch->fMaxData){
      ch->fMaxData = abs(data)/ch->fSDev;
      ch->fMaxTime = currenttime;
      ch->fMaxSDev = ch->fSDev;
      ch->fMaxMean = ch->fMean;
      }

    }

  }
  else{  // data point below threshold

    if (ch->fEventStatus == ChannelStatus::kEvent) { 
      ch->fEndTime     = currenttime;	
      ch->fEventStatus = ChannelStatus::kWait;     
    } 
    else if (ch->fEventStatus == ChannelStatus::kWait){

      if (currenttime - ch->fEndTime > Interval(kWaitTime) ){

	// always send trigger for high threshold
	if (ch->fMaxData > ch->fThresholdHi){
	  OutputEvent(ch);
	  ch->fEventCount++;

	  // for low threshold, check event duration and data point density
	}
	else if (ch->GetEventDuration() > kMinEventDuration && 
		 ch->fEventNPoint > 
		 ch->GetEventDuration() * kMinDataPointDensity *
		 ch->fSampleFreq){
	  OutputEvent(ch);
	  ch->fEventCount++;
	}

	// clear 
	ch->fEventNPoint = 0; 
	ch->fEventStatus = ChannelStatus::kNone;
      }
    }
  }
}


//output glitchdata data
void magGlitch::OutputEvent(ChannelStatus* ch)
{

  fDataFile<<ch->fName<<" "
	   <<setprecision(15)<<ch->fStartTime<<" "
	   <<setprecision(15)<<ch->fEndTime<<" "
	   <<setprecision(15)<<ch->fMaxTime<<" "
	   <<setprecision(6) <<ch->fMaxData<<" "
	   <<ch->fMaxMean<<" "
	   <<ch->fMaxSDev<<" "
	   <<ch->fEventNPoint<<" "
	   <<ch->fTrigComment
	   <<endl;
  ch->fHistCur->Fill(ch->fMaxData);
  ch->fHistTot->Fill(ch->fMaxData);
  
  //send trigger to Trigger Manager if triggerOn is true
  if (fTriggerON){
    string subID;
    if ( ch->fTrigComment.size() ) subID = ch->fTrigComment;
    else subID = ch->fName;

    trig::TrigRslt trigData("magGlitch",subID.c_str());

    trigData.setTime(ch->fStartTime);
    trigData.setDuration(ch->fEndTime - ch->fStartTime);
    trigData.setIntensity(ch->fMaxData);
    trigData.setPriority(trig::p_info);
    trigData.setDisposition(trig::d_metaDB);
    int trigerr = sendTrigger(trigData);
    if (trigerr) {
      fLogFile<<"Error Sending Trigger "<<trigerr<<" "
	      <<ch->fEndTime<<endl;
    }
  }

  return;
}


// output log file 
void magGlitch::OutputLog()
{

  long currenttime = long(fCHStatus.begin()->fTSnew->getEndTime().getS());
  if (!fLogStartTime) { // output magGlitch start time
    fLogStartTime = long(fCHStatus.begin()->fTSnew->getStartTime().getS());
    fLogFile << "Monitor Start = "
	     << fLogStartTime 
	     << endl
	     <<"------------------------------------"
	     << endl;
  }
	
  // update log file every fLogInterval
  if (currenttime - fLogStartTime >= fLogInterval){
    fLogFile<<"Log Start: "<<fLogStartTime
	    <<"\tInterval "<<fCount/fLogInterval<<": "<<fLogInterval<<" sec"
	    <<endl;
    
    for (vector<ChannelStatus>::iterator ch = fCHStatus.begin();
	 ch != fCHStatus.end(); ch++) { // output channel statistics
      fLogFile<<ch->fName<<" ";
      if (ch->fCheckEventCount){ // check # of CheckEvent executed in fLogInterval
	fLogFile<<ch->fEventCount<<" "
		<<ch->fSumMean/ch->fCheckEventCount<<" ";
	if (ch->fSumSDevCount) {
	  fLogFile<<ch->fSumSDev/ch->fSumSDevCount<<" ";
	}
	else { 
	  fLogFile<<-1<<" ";
	}
	fLogFile<<(float)ch->fCheckEventCount/fLogInterval
		<<endl;
      }
      else { // # of main loops executed == 0
	fLogFile<<-1
		<<endl;
      }
      // next, output all the histogram info
      ch->fHistCur->Dump(fDataFile);
      fDataFile<<endl;
      ch->ResetHist(Time(currenttime,0));
      string temp = string(ch->fHistCur->GetNLabel()) + " "
	+ tostring(fCount/fLogInterval)	+ " interval(s)";
      ch->fHistTot->SetNLabel(temp.c_str());
      ch->fSumMean = 0;
      ch->fCheckEventCount = 0;
      ch->fEventCount = 0;
      ch->fSumSDev = 0;
      ch->fSumSDevCount = 0;
    }
    fLogFile<<"------------------------------------"<<endl;
    fLogStartTime = currenttime;
  }

  return;  

}

void magGlitch::Reset()
{
  if (fData) delete fData;
  fData = 0;
  if (fFilterData) delete fFilterData;
  fFilterData = 0;
  fMaxData = fMinData = 0;
  fNChannel = 0;
  fDAQInterval = 0;
  fLogInterval = 0;
  fDecayTime = fDecay = 0;
  fCount = 0;
  fLastEndTime = 0;
  fLogStartTime = 0;
  fLogName = "";
  fDataName = "";
  fTriggerON = 0;
}

const float magGlitch::GetMean(const ChannelStatus* ch) const
{

  int step;

  switch (ch->fSampleFreq){
  case 2048:
    step=8;
    break;
  case 16384:
    step=64;
    break;
  default:
    step=1;
  }
  
  int n = 0;
  int outn = 0;
  float avg = 0;
  float outavg = 0;
  // use last 1-sec of 2-sec data
  const float* data = fData + ch->fSampleFreq;

  if (ch->fInitCount){ 
    for (int i=0; i < ch->fSampleFreq; i+=step){
      if (abs(data[i]) < 32768){
	avg = avg + data[i];
	n++;
      }
      else {
	outavg = outavg + data[i];
	outn++;
      }
    }
  }
  else {
    for (int i=0; i < ch->fSampleFreq; i+=step){
      if ((abs(data[i] - ch->fMean) < 2 * ch->fThresholdHi * ch->fSDev) && 
	  abs(data[i])<32768 ){
	avg = avg + data[i];
	n++;
      }
      else {
	outavg = outavg + data[i];
	outn++;
      }
    }
  }

  if (n) return avg/n;
  else return outavg/outn;
    
}

const float magGlitch::GetSDev(const int sfreq) const
{

  int step;
  switch (sfreq){
  case 2048:
    step=8;
    break;
  case 16384:
    step=64;
    break;
  default:
    step=1;
  }

  int n = 0;
  float sdev = 0;

  for (int i = sfreq * (kOffsetFactor - 1); 
       i < sfreq * kOffsetFactor; i += step){
    if (abs(fData[i]) < 32768 ){
      sdev = sdev + fData[i] * fData[i];
      n++;
    }
  }

  if ( n ) sdev = sqrt(sdev/n);
  else sdev = -1;

  return sdev;
}

const float magGlitch::GetSDevF(const int sfreq) const
{
  int step;
  switch (sfreq){
  case 2048:
    step=8;
    break;
  case 16384:
    step=64;
    break;
  default:
    step=1;
  }
  int n = 0;
  float sdev = 0;

  for (int i = sfreq * (kOffsetFactor - 1);
       i < sfreq * kOffsetFactor; i += step){
    if (abs(fData[i]) < 32768 ){
      sdev = sdev + fFilterData[i] * fFilterData[i];
      n++;
    }
  }
  if ( n ) sdev = sqrt(sdev/n);
  else sdev = -1;

  return sdev;
}

void magGlitch::UpdateSDevF(const float sdev, ChannelStatus* ch)
{
  float norm = 1 + fDecay * ch->fLastNorm;
  ch->fSDev = (sdev + ch->fSDev * fDecay * ch->fLastNorm)/norm;
  ch->fLastNorm = norm;
}

// Reset the Histogram information
void ChannelStatus::ResetHist(const Time &time) {
    fHistCur->Reset();
    // always have 20 bins spaced .5 sigma apart, minimum at an integer.
    double histmin = floor(fThresholdLow);
    int histbin(kHistBinNumber);
    double histmax = histmin + kHistBinSize * histbin;
    string histname = fName + ", " + tostring(fCutOffFreqLow)
      + "-" + tostring(fCutOffFreqHi) + "Hz";
    string histxlbl = "Triggers for " + histname + " (std. dev.)";
    string histylbl = "# of events";
    fHistCur->SetTitle(histname.c_str());
    fHistCur->SetXLabel(histxlbl.c_str());
    fHistCur->SetNLabel(histylbl.c_str());
    fHistCur->SetBinLowEdges(histbin,histmin,histmax);
    fHistCur->SetTime(time);
}

//'tostring' changes an integer into a string
template<class T>
string tostring(const T& n) {
  ostringstream ost;
  ost << n << ends;
  return ost.str();
}

//'shift' shifts a time series by amount dt, or the nearest whole data point
//  bool shift(TSeries &ts, const Interval &dt) {
//    DVector::size_t index = (DVector::size_t)floor(dt / ts.getTStep());
//    DVector::size_t length = (DVector::size_t)floor( (ts.getInterval() - dt)/
//  						   ts.getTStep() + .5       );
//    const DVector* temp1 = ts.refDVect();
//    DVector *temp2 = new DVector( *(temp1->Extract(index,length)) );
//    if (temp->s_data()) short *tdata = new short[(int)temp->getLength()];
//    else if (temp->I_data()) int *tdata = new int[(int)temp->getLength()];
//    else if (temp->F_data()) float *tdata = new float[(int)temp->getLength()];
//    else if (temp->D_data()) double *tdata = new double[(int)temp->getLength()];
//    else if (temp->C_data()) 
//      fComplex *tdata = new fComplex[(int)temp->getLength()];
//    else {
//      delete temp;
//      return 0;
//    }
//    Interval dt2(ts.getTStep());
//    dt2 *= double(index);
//    Time tnext = ts.getStartTime() + dt2;
//    ts.setData(tnext, ts.getTStep(), tdata, (TSeries::size_t)temp->getLength());
//    delete temp;
//    delete tdata;
//    return ts.isGood();
//  }			     
