#ifndef _CHANNEL_H_

#define _CHANNEL_H_

// standard
#include <vector>
#include <list>

// DMT
#include <FSeries.hh>
#include <Complex.hh>
#include <IIRFilter.hh>

typedef fComplex Complex;

typedef list<Complex> ListComplex;

class ChnInfo { // description of a channel

  bool OSC; // set to 1 is sensitive to OSC

  unsigned int cid; // channel Id

  string name; // names of channel

  unsigned int fs; // sampling freq (after decimation)

  unsigned int dec; // decimation factor

  Interval timeRes; // time resolution of spectrogram


  double fLow,
    fHigh;       // bound on freq in Hz for running clustering analysis


  Interval stabScale; // time scale for background noise stability

  double stabTol; // tolerance on background noise stability

  vector<unsigned int> testFreq; // indices of test frequencies

  vector<ListComplex*> History;  // past values of real and imaginary components

  // buffers to compute mean and variance
  unsigned int Nsum;
  double *mr, *mi, *s2r, *s2i;

  bool ForceComputeThreshold; // set to 1 to force recomputation
  bool thrCompMode; // set to 1 during acquisition of data for thresholds comp

  Complex* MeanValues; // mean Fourier components vs frequency; of size fs
  Complex* VarValues; // variances of the real and imaginary parts



  double bpp; // black pixel probability

  vector<float> powThresh; // thresholds on power vs frequency


  unsigned int sigma; // minimum cluster size

  vector<unsigned int> delta; // minimum distances


  double alpha; // confidence level for power test


  EventsList* eventsList; // list of events for that channel


  bool doFilter; // set to 1 if need to apply a filter
  vector<dComplex> poles, zeroes;
  IIRFilter *filter;

 public:

  ChnInfo (const int cid_,
	   const string& name_,
	   const unsigned int fs_,
	   const double fLow_,
	   const double fHigh_,
	   const unsigned int dec_,
	   const Interval timeRes_,
	   const Interval stabScale_,
	   const double stabTol_,
	   const vector<unsigned int>& testFreq_,
	   const double p_,
	   const unsigned int sigma_,
	   const vector<unsigned int>& delta_,
	   const double alpha_);    // Constructor when thresholds are to be evaluated from the data


  ChnInfo (const int cid_,
	   const string& name_,
	   const unsigned int fs_,
	   const double fLow_,
	   const double fHigh_,
	   const unsigned int dec_,
	   const Interval timeRes_,
	   const Interval stabScale_,
	   const double stabTol_,
	   const vector<unsigned int>& testFreq_,
	   const double p_,
	   const unsigned int sigma_,
	   const vector<unsigned int>& delta_,
	   const double alpha_,
	   const vector<float>& powThresh);    // Constructor when thresholds are known


  void ProcessData(unsigned int cid_,
		   const FSeries* s); // process one spectrum; cid_ should match cid


  // NOTE: at this time, TestStationarity will test stationarity using
  // a chi-2 test every stabScale seconds. If not Stationary, ComputeThresholds
  // recomputes the thresholds and mean values OVER A TIMESCALE of stabScale
  // also.
  bool TestStationarity (); // test background noise for stationarity

  void ComputeThresholds (const FSeries *); // compute the thresholds

  ~ChnInfo ();   // Destructor

  // inline functions
  void SetFs (unsigned int nfs) { fs = nfs; }
  unsigned int Getcid () {return cid;}
  unsigned int GetFs () { return fs;}
  double GetTimeRes () { return (double)timeRes; }
  double GetStabTol () { return stabTol; }
  double GetStabTimeScale () { return (double)stabScale; }
  vector<unsigned int>& GetTestFreq () { return testFreq; }
  string& GetName () { return name; }
  double GetfLow () { return fLow; }
  double GetfHigh () { return fHigh; }
  unsigned int GetSigma () { return sigma; }
  double GetAlpha () { return alpha; }
  double GetBPP () { return bpp; }
  vector<unsigned int>& GetDelta () {return delta;}
  void ComputeThreshold () { ForceComputeThreshold = 1; Nsum = 0;}
  void SetBPP(double b) {bpp=b; ComputeThreshold();}
  bool DoOSC () {return OSC;}
  void SetOSC (bool b) {OSC=b;}
  void ScaleThreshold (float sf);

  bool DoFilter () {return doFilter;}
  void ConstructFilter();
  void SetPoles(vector<dComplex> &p) {doFilter=1; poles = p;}
  void SetZeroes(vector<dComplex> &p) {doFilter=1; zeroes = p;}
  IIRFilter * GetFilter() {return filter;}
};  




#endif
