#ifndef TRENDCHAN_HH
#define TRENDCHAN_HH

#include <string>
#include "TSeries.hh"
#include "TrendAcc.hh"

class FrWriter;

/**  Accumulate and write trend data. The trend is given  minor and a 
  *  major interval. The trend averages over all samples within the minor
  *  interval and produces a frame over the major interval.
  *  @memo Make trend frames.
  *  @author John Zweizig.
  *  @version 1.0; Modified February 28, 2000.
  */
class TrendChan {
public:
  /**  Data type for data counts.
    */
  typedef TrendAcc::count_t count_t;

  /**  Arithmetic data type.
    */
  typedef TrendAcc::math_t math_t;

  /**  TSeries data type.
    */
  typedef float trend_t;

public:
  /**  Check that the specified string is a valid DMT trend channel name.
    *  @memo Check for valid name.
    *  @param chan Channel name string to be tested
    *  @return true if name is valid.
    */
  static bool validName(const std::string& chan, std::string& monid);

public:
  /**  Default Constructor
   */
  TrendChan(void);

  /**  Copy Constructor
   */
  TrendChan(const TrendChan& x);
  /**  Constructor
   */
  explicit TrendChan(const char* Name, Interval nAvg);

  /**  Destructor
   */
  ~TrendChan();

  /** Assignment operator
   */
  TrendChan& operator=(const TrendChan& x);

  /**  Return the base name of the trend channel time series to be generated.
    */
  std::string getName(void) const;

  /**  Return the end of the time series.
    */
  Time getEndTime(void) const;

  /**  Get the number of samples in the current frame.
    */
  long getNSample(const Time& tStart, const Time& tStop) const;

  /**  Test whether any data has been trended.
    *  @memo Test for trended data.
    *  @return True if trend data exist.
    */
  bool isEmpty(void) const;

  /**  Add a data point to the trend running average.
    *  @memo add a single point.
    *  @param point The value of the next element of a series.
    */
  void addData(const TSeries& ts);

  /**  Add a data point to the trend running average.
    *  @memo add a single point.
    *  @param point The value of the next element of a series.
    */
  void addData(const Time& t, math_t point);

  /**  Delete data up to (but not including) the specified time.
    *  Data are discarded from all time series in the trend channel.
    *  @memo remove data from front of trend.
    *  @param t new trend start-time.
    */
  void discard(const Time& t);

  /**  Extend all the trend series to the specified time.
    *  @memo Extend time series.
    *  @param t new trend end-time.
    */
  void extend(const Time& t);

  /**  Get a reference to the current average value series
    */
  const TSeries& refAvgSeries(void) const;

  /**  Get a reference to the current maximum entry series
    */
  const TSeries& refMaxSeries(void) const;

  /**  Get a reference to the current minimum entry series
    */
  const TSeries& refMinSeries(void) const;

  /**  Get a reference to the current entry count series
    */
  const TSeries& refNSeries(void) const;

  /**  Get a reference to the current RMS series
    */
  const TSeries& refRmsSeries(void) const;

  /**  Set the portion of the current trend covered by the argument series
    *  to the argument data. All data (mean, rms, min, max) are set to the
    *  same values and the number of averages is set to 1 over the same range.
    */
  void setData(const TSeries& Avg);

  /**  Set the portion of the current trend covered by the argument series
    *  to the argument data.
    */
  void setData(const TSeries& Avg, const TSeries& Rms, const TSeries& Min, 
	       const TSeries& Max, const TSeries& Num);

  /**  Set the trend name and assign names to the constituent time series.
    *  When the trend is written these derived names will be used as the
    *  "ADC" names in the frame file.
    *  @memo Set the trend name.
    *  @param Name Name for the trend data.
    */
  void setName(const char* Name);

  /**  Start accumulating a trend series. Any data alrady accumulated before 
    *  the frame start time is discarded. The frame start time is saved.
    *  @memo Start a new trend frame.
    */
  void startFrame(const Time& start);

  /**  Force current point into trend series, and pad the time series to 
    *  the specified time.
    *  @memo Synchronize trend series to current data.
    *  @param t Time to which the data will be padded.
    */
  void synch(const Time& t=Time(0));

  /**  Write data to a frame.
    */
  void write(FrWriter& writer);

private:
  /**  Reset the running average.
    *  @memo add a single point.
    *  @param point The value of the next element of a series.
    */
  void reset(void);

  /**  Start accumulating a new trend point at the specified time. The 
    *  accumulation start time is moved forward to the nearest point time
    *  boundary.
    *  @memo Start to accumulate a new trend point.
    *  @param start a new trend point.
    */
  void startAcc(const Time& start);

  /**  Add the current running average to the time series.
    *  @memo add a single point.
    *  @param point The value of the next element of a series.
    */
  void trendPoint(void);

  /**  Append the current running average to the end of the time series.
    *  @memo add a single point.
    *  @param point The value of the next element of a series.
    */
  void appendPoint(void);

  /**  Add the current running average to an appropriate point in the
    *  time series.
    *  @memo add a single point.
    *  @param point The value of the next element of a series.
    */
  void incrementPoint(void);

private:
  /**  Trend channel name.
    */
  std::string mName;

  /**  Running total of accumulated points.
    */
  TrendAcc mAccum;

  /**  Start time for the current frame.
    */
  Time mStartFrame;

  /**  Start time for current accumulation period.
    */
  Time mStartAcc;

  /**  Accumulation interval.
    */
  Interval mAccIntvl;

  /**  Trend data - Number of samples.
    */
  TSeries mNSampTS;

  /**  Trend data - Average value.
    */
  TSeries  mAvgTS;

  /**  Trend data - RMS value.
    */
  TSeries  mRmsTS;

  /**  Trend data - Minimum Value.
    */
  TSeries  mMinTS;

  /**  Trend data - Maximum value.
    */
  TSeries  mMaxTS;
};

//======================================  Inline functions
#ifndef __CINT__

inline bool
TrendChan::isEmpty(void) const {
    return !mStartAcc && mNSampTS.isEmpty();
}

inline std::string 
TrendChan::getName(void) const {
    return mName;
}

inline const TSeries& 
TrendChan::refAvgSeries(void) const {
    return mAvgTS;
}

inline const TSeries& 
TrendChan::refMaxSeries(void) const {
    return mMaxTS;
}

inline const TSeries& 
TrendChan::refMinSeries(void) const {
    return mMinTS;
}

inline const TSeries& 
TrendChan::refNSeries(void) const {
    return mNSampTS;
}

inline const TSeries& 
TrendChan::refRmsSeries(void) const {
    return mRmsTS;
}
#endif  // def(__CINT__)

#endif  // TRENDCHAN_HH
