#ifndef MULTIPIPE_HH
#define MULTIPIPE_HH

#include "Pipe.hh"
#include <vector>

/**  MultiPipe is an arbitrary compound filter composed of one or more
  *  Pipe segments. 
  *  @memo Compoud filter class
  *  @author John G. Zweizig
  *  @version 1.1; Last modified December 11, 2001
  */
class MultiPipe : public Pipe {
public:
   using Pipe::apply;
   using Pipe::dataCheck;
   
  /// Pipe array type
  typedef std::vector<Pipe*> PipeConfig;

  /**  Construct an empty MultiPipe.
    *  @memo Default constructor
    */
  MultiPipe(void);

  /**  Construct a MultiPipe identical to an existing argument.
    *  @memo Copy constructor
    */
  MultiPipe(const MultiPipe& p);

  /**  Destroy a MultiPipe.
    *  @memo Destructor
    */
  ~MultiPipe(void);

  /**  Create a copy of the MultiPipe.
    *  @memo Clone a MultiPipe.
    */
  MultiPipe* clone(void) const;

  /**  Assignment.
    *  @memo Assignment operator
    */
  MultiPipe& operator= (const MultiPipe& p);
  
  /**  Add a filter to the MultiPipe. 
    *  @memo Add a filter.
    */
  Pipe* addPipe(const Pipe& p);
  
  /**  Return the filter array. 
    *  @memo Return filter array.
    */
  const PipeConfig& pipe() const {
     return mPipe; }

  /**  Apply the compound filter to a time series. Each filter is applied 
    *  to the time series in the order that it was added to the MultiPipe.
    *  @memo Filter a time series.
    *  @param ts Input time series.
    *  @return Filtered time series.
    */
  TSeries apply(const TSeries& ts);

  /**  Check that the specified time series is valid input for the 
    *  compound filter. An exception is thrown if the MultiPipe can not
    *  be applied to the specified data.
    *  @memo Check data is valid for filtering.
    *  @param ts Time series to be checked for validity. 
    */
  void dataCheck(const TSeries& ts) const;

  /**  Test whether the compound filter is in use. The compound filter is
    *  considered to be in use if any of the component filters is in use.
    *  @memo Test whether filter is in use.
    *  @return True if compound filter is in use.
    */
  bool inUse(void) const;

  /**  Get start time of the data segment currently in progress.
    *  @memo Get start time of epoch.
    *  @return Start time.
    */
  Time getStartTime(void) const;

  /**  Get the Time of the next expected sample to be processed.
    *  @memo Get current time.
    *  @return Current time.
    */
  Time getCurrentTime(void) const;

  /**  Get the additional gain.
    *  @memo Get gain.
    *  @return gain.
    */
  double getGain(void) const;


  /**  Get the delay time of the combined filters.
    *  @memo Get time delay.
    *  @return Interval delay.
    */
  Interval getTimeDelay(void) const;

  /**  Set the additional gain.
    *  @memo Set gain.
    *  @param Gain.
    *  @return void
    */
  void setGain(double gain); 
  
  /**  Reset all filters.
    *  @memo Reset
    */
  void reset();

protected:
  /** The transfer coefficient of the filter at the specified 
    * frequency is calculated and returned as a complex number. 
    * @memo Get a transfer coefficent of a Filter.
    * @param coeff a complex number representing the Filter response 
    *              at the specified frequency (return)
    * @param f Frequency at which to sample the transfer function.
    * @return true if successful       
    */
  bool xfer (fComplex& coeff, double f) const throw();
		     
private:
  /// Pipe array
  PipeConfig mPipe;
  /// Additional Gain factor
  double mGain;
};

#endif  // MULTIFILTER_HH
